Package: eppoFindeR
Type: Package
Title: Interface to the EPPO Database and Public APIs
Version: 2.0.0
Maintainer: Luca Belmonte <luca.belmonte@efsa.europa.eu>
Authors@R: 
  c(person(given = "Lorenzo",
           family = "Copelli",
           role = "aut",
           comment = c(ORCID = "0009-0002-4305-065X")),
    person(given = "Dayana Stephanie",
           family = "Buzle",
           role = "aut",
           comment = c(ORCID = "0009-0003-2990-7431")),
    person(given = "Rafael",
           family = "Vieira",
           role = "aut",
           comment = c(ORCID = "0009-0009-0289-5438")),
    person(given = "Agata",
           family = "Kaczmarek",
           role = "aut",
           comment = c(ORCID = "0000-0002-7463-5821")),
    person(given = "Luca",
           family = "Belmonte",
           role = c("aut", "cre"),
           email = "luca.belmonte@efsa.europa.eu",
           comment = c(ORCID = "0000-0002-7977-9170")))
Description: Provides an interface to the public APIs of the European and
    Mediterranean Plant Protection Organization (EPPO) database. It enables
    users to retrieve EPPO data by accessing specific services and datasets. The
    package also includes utilities for data wrangling, including the
    integration of taxonomy with rank information. See
    <https://data.eppo.int/apis/> for more details.
License: EUPL-1.2
URL: https://openefsa.github.io/eppoFindeR/
BugReports: https://github.com/openefsa/eppoFindeR/issues
Depends: R (>= 4.1.0)
Imports: cli (>= 3.6.5), httr2 (>= 1.2.1), jsonlite (>= 1.8.8), dplyr
        (>= 1.1.4), glue (>= 1.7.0), purrr (>= 1.0.2), checkmate (>=
        2.3.1), tibble(>= 3.3.0)
Suggests: devtools (>= 2.4.5), roxygen2 (>= 7.3.1), testthat (>=
        3.2.1), usethis (>= 2.2.3), knitr (>= 1.0), rmarkdown (>= 2.0),
        covr (>= 3.6.4)
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/Needs/website: pkgdown
Repository: CRAN
NeedsCompilation: no
Packaged: 2026-04-22 08:06:59 UTC; copello
Author: Lorenzo Copelli [aut] (ORCID: <https://orcid.org/0009-0002-4305-065X>),
  Dayana Stephanie Buzle [aut] (ORCID:
    <https://orcid.org/0009-0003-2990-7431>),
  Rafael Vieira [aut] (ORCID: <https://orcid.org/0009-0009-0289-5438>),
  Agata Kaczmarek [aut] (ORCID: <https://orcid.org/0000-0002-7463-5821>),
  Luca Belmonte [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7977-9170>)
Date/Publication: 2026-04-24 18:10:02 UTC
Built: R 4.6.0; ; 2026-04-26 12:55:50 UTC; unix

Package: eppoFindeR
Type: Package
Title: Interface to the EPPO Database and Public APIs
Version: 2.0.0
Maintainer: Luca Belmonte <luca.belmonte@efsa.europa.eu>
Authors@R: 
  c(person(given = "Lorenzo",
           family = "Copelli",
           role = "aut",
           comment = c(ORCID = "0009-0002-4305-065X")),
    person(given = "Dayana Stephanie",
           family = "Buzle",
           role = "aut",
           comment = c(ORCID = "0009-0003-2990-7431")),
    person(given = "Rafael",
           family = "Vieira",
           role = "aut",
           comment = c(ORCID = "0009-0009-0289-5438")),
    person(given = "Agata",
           family = "Kaczmarek",
           role = "aut",
           comment = c(ORCID = "0000-0002-7463-5821")),
    person(given = "Luca",
           family = "Belmonte",
           role = c("aut", "cre"),
           email = "luca.belmonte@efsa.europa.eu",
           comment = c(ORCID = "0000-0002-7977-9170")))
Description: Provides an interface to the public APIs of the European and
    Mediterranean Plant Protection Organization (EPPO) database. It enables
    users to retrieve EPPO data by accessing specific services and datasets. The
    package also includes utilities for data wrangling, including the
    integration of taxonomy with rank information. See
    <https://data.eppo.int/apis/> for more details.
License: EUPL-1.2
URL: https://openefsa.github.io/eppoFindeR/
BugReports: https://github.com/openefsa/eppoFindeR/issues
Depends: R (>= 4.1.0)
Imports: cli (>= 3.6.5), httr2 (>= 1.2.1), jsonlite (>= 1.8.8), dplyr
        (>= 1.1.4), glue (>= 1.7.0), purrr (>= 1.0.2), checkmate (>=
        2.3.1), tibble(>= 3.3.0)
Suggests: devtools (>= 2.4.5), roxygen2 (>= 7.3.1), testthat (>=
        3.2.1), usethis (>= 2.2.3), knitr (>= 1.0), rmarkdown (>= 2.0),
        covr (>= 3.6.4)
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/Needs/website: pkgdown
Repository: CRAN
NeedsCompilation: no
Packaged: 2026-04-22 08:06:59 UTC; copello
Author: Lorenzo Copelli [aut] (ORCID: <https://orcid.org/0009-0002-4305-065X>),
  Dayana Stephanie Buzle [aut] (ORCID:
    <https://orcid.org/0009-0003-2990-7431>),
  Rafael Vieira [aut] (ORCID: <https://orcid.org/0009-0009-0289-5438>),
  Agata Kaczmarek [aut] (ORCID: <https://orcid.org/0000-0002-7463-5821>),
  Luca Belmonte [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7977-9170>)
Date/Publication: 2026-04-24 18:10:02 UTC
Built: R 4.5.2; ; 2026-04-24 20:45:16 UTC; unix

# Generated by roxygen2: do not edit by hand

export(country)
export(general)
export(references)
export(reportingService)
export(rppo)
export(taxon)
export(taxons)
export(tools)
export(uniformTaxonomy)
importFrom(checkmate,assert)
importFrom(checkmate,assert_class)
importFrom(checkmate,assert_data_frame)
importFrom(checkmate,assert_list)
importFrom(checkmate,assert_names)
importFrom(checkmate,assert_string)
importFrom(checkmate,assert_true)
importFrom(checkmate,assert_vector)
importFrom(checkmate,check_data_frame)
importFrom(checkmate,check_list)
importFrom(cli,cli_abort)
importFrom(dplyr,any_of)
importFrom(dplyr,bind_rows)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,relocate)
importFrom(dplyr,select)
importFrom(glue,glue)
importFrom(httr2,req_error)
importFrom(httr2,req_headers)
importFrom(httr2,req_method)
importFrom(httr2,req_perform)
importFrom(httr2,req_url_query)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(httr2,resp_body_string)
importFrom(httr2,resp_status)
importFrom(jsonlite,fromJSON)
importFrom(purrr,imap)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,set_names)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)

## ----include=FALSE------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# install.packages("eppoFindeR")

## ----eval=FALSE---------------------------------------------------------------
# # install.packages("devtools")
# devtools::install_github("openefsa/eppoFindeR")

## ----apiKeyExample, eval=FALSE------------------------------------------------
# taxon(c("BEMITA"), apiKey = "<your_eppo_api_key>")

## ----loadLibrary, eval=FALSE--------------------------------------------------
# library(eppoFindeR)

## ----eval=FALSE---------------------------------------------------------------
# help("<function_name>")

## ----eval=FALSE---------------------------------------------------------------
# help("taxon")

## ----taxonAllServices, eval=FALSE---------------------------------------------
# taxon(c("BEMITA"))

## ----toolsName2Codes, eval=FALSE----------------------------------------------
# tools(
#   c("name2codes"), # The service to query (/tools/name2codes).
#   params = list(
#     name2codes = list(
#       name = "Bemisia tabaci" # The required parameter.
#     )
#   )
# )

## ----queryService, eval=FALSE-------------------------------------------------
# taxon(c("BEMITA"), c("categorization"))

## ----queryServiceOutput, eval=FALSE-------------------------------------------
# $categorization
# # A tibble: 20 × 12
#    queried_eppo_code continent_id continent_name country_iso country_name    qlist qlist_label year_add
#    <chr>                    <int> <chr>          <chr>       <chr>           <chr> <chr>          <int>
#  1 BEMITA                       4 America        CL          Chile           1     A1 list         2019
#  2 BEMITA                       2 Asia           BH          Bahrain         2     A2 list         2003
#  3 BEMITA                       2 Asia           KZ          Kazakhstan      1     A1 list         2017
#  4 BEMITA                       1 Europe         AZ          Azerbaijan      1     A1 list         2024
#  5 BEMITA                       1 Europe         BY          Belarus         X     Quarantine…     1994
#  6 BEMITA                       1 Europe         GE          Georgia         1     A1 list         2018
#  7 BEMITA                       1 Europe         MD          Moldova, Repub… X     Quarantine…     2017
#  8 BEMITA                       1 Europe         NO          Norway          X     Quarantine…     2012
#  9 BEMITA                       1 Europe         RU          Russian Federa… 2     A2 list         2014
# 10 BEMITA                       1 Europe         RS          Serbia          1     A1 list         2015
# 11 BEMITA                       1 Europe         CH          Switzerland     1     A1 list         2019
# 12 BEMITA                       1 Europe         TR          Türkiye         2     A2 list         2016
# 13 BEMITA                       1 Europe         UA          Ukraine         1     A1 list         2019
# 14 BEMITA                       1 Europe         GB          United Kingdom  1     A1 list         2020
# 15 BEMITA                       5 Oceania        NZ          New Zealand     X     Quarantine…     2000
# 16 BEMITA                       6 RPPO/EU        9M          EAEU            2     A2 list         2016
# 17 BEMITA                       6 RPPO/EU        9A          EPPO            2     A2 list         1989
# 18 BEMITA                       6 RPPO/EU        9L          EU              P     PZ Quarant…     2019
# 19 BEMITA                       6 RPPO/EU        9L          EU              N     A1 Quarant…     2019
# 20 BEMITA                       6 RPPO/EU        9H          OIRSA           2     A2 list         1992
# # ℹ 4 more variables: year_delete <lgl>, year_transient <lgl>, queried_url <chr>, queried_on <dttm>

## ----queryWhole, eval=FALSE---------------------------------------------------
# data <- taxon(c("BEMITA"))

## ----queryAllOutput, eval=FALSE-----------------------------------------------
# # Print the names of the services in the result.
# names(data)
#  [1] "overview"           "infos"              "names"              "taxonomy"
#  [5] "categorization"     "kingdom"            "hosts"              "pests"
#  [9] "vectors"            "vectorof"           "bca"                "bcaof"
# [13] "photos"             "reporting_articles" "documents"          "standards"
# [17] "distribution"
# 
# # Print the data from a specific service.
# data$overview
# # A tibble: 1 × 11
#   queried_eppo_code eppocode datecreate         lastupdate infos replacedby prefname is_active datatype
#   <chr>             <chr>    <chr>              <chr>      <lgl> <lgl>      <chr>    <lgl>     <chr>
# 1 BEMITA            BEMITA   2002-10-28 00:00:… 2002-10-2… NA    NA         Bemisia… TRUE      GAI
# # ℹ 2 more variables: queried_url <chr>, queried_on <dttm>

## ----eval=FALSE---------------------------------------------------------------
# data <- taxon(c("BEMITA"), c("taxonomy"))

## ----uniformTaxonomy, eval=FALSE----------------------------------------------
# taxonomy <- uniformTaxonomy(data$taxonomy)

## ----uniformTaxonomyOutput, eval=FALSE----------------------------------------
# # A tibble: 11 × 6
#    type      queried_eppo_code eppocode prefname       queried_url                  queried_on
#    <chr>     <chr>             <chr>    <chr>          <chr>                        <dttm>
#  1 Kingdom   BEMITA            1ANIMK   Animalia       https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  2 Phylum    BEMITA            1ARTHP   Arthropoda     https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  3 Subphylum BEMITA            1HEXAQ   Hexapoda       https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  4 Class     BEMITA            1INSEC   Insecta        https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  5 Subclass  NA                NA       NA             NA                           NA
#  6 Order     BEMITA            1HEMIO   Hemiptera      https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  7 Suborder  BEMITA            1STERR   Sternorrhyncha https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  8 Family    BEMITA            1ALEYF   Aleyrodidae    https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
#  9 Subfamily NA                NA       NA             NA                           NA
# 10 Genus     BEMITA            1BEMIG   Bemisia        https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
# 11 Species   BEMITA            BEMITA   Bemisia tabaci https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50

## ----multipleEPPOCodes, eval=FALSE--------------------------------------------
# data <- taxon(c("BEMITA", "GOSHI"))

## ----multipleEPPOCodesOutput, eval=FALSE--------------------------------------
# # Print the names of the services in the result.
# names(data)
#  [1] "overview"           "infos"              "names"              "taxonomy"
#  [5] "categorization"     "kingdom"            "hosts"              "pests"
#  [9] "vectors"            "vectorof"           "bca"                "bcaof"
# [13] "photos"             "reporting_articles" "documents"          "standards"
# [17] "distribution"
# 
# # Print the data from a specific service.
# data$overview
# # A tibble: 2 × 11
#   queried_eppo_code eppocode datecreate         lastupdate infos replacedby prefname is_active datatype
#   <chr>             <chr>    <chr>              <chr>      <chr> <lgl>      <chr>    <lgl>     <chr>
# 1 BEMITA            BEMITA   2002-10-28 00:00:… 2002-10-2… NA    NA         Bemisia… TRUE      GAI
# 2 GOSHI             GOSHI    2004-04-21 00:00:… 2004-04-2… <p>M… NA         Gossypi… TRUE      PFL
# # ℹ 2 more variables: queried_url <chr>, queried_on <dttm>

#' Query the EPPO API RPPO endpoints.
#' 
#' This function queries the RPPO endpoints of the EPPO Global Database via
#' REST API for one or more RPPO code(s) and one or more service(s). For each
#' RPPO code in `rppoCodes`, the function sequentially queries all specified
#' `services` and returns the extracted data through a list of dataframes.
#' 
#' @param rppoCodes `character` (vector). One or more RPPO codes to query.
#'   Available RPPO codes can be retrieved via the `references/rppos` service.
#'
#' @param services `character` (vector). One or more RPPO services to query.
#'   Common service names include: `overview` and `categorization`. A
#'   validation step ensures that all provided services match the supported
#'   service names.
#' 
#'   By default: all services.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content for all the specified RPPO
#'   codes.
#' 
#' @importFrom checkmate assert_vector assert_string
#' @importFrom purrr map set_names
#'
#' @examples
#' \dontrun{
#' # Get all information about the European and Mediterranean Plant Protection
#' # Organisation.
#' rppoData_ <- rppo(rppoCodes = c("9A"))
#' 
#' # Get categorization data about the European and Mediterranean Plant
#' # Protection Organisation.
#' rppoData_ <- rppo(rppoCodes = c("9A"), services = c("categorization"))
#' 
#' # Get basic data about European and Mediterranean Plant Protection
#' # Organisation and the European Union.
#' rppoData_ <- rppo(rppoCodes = c("9A", "9L"), services = c("overview"))
#' }
#' 
#' @export
#' 
rppo <- function(
  rppoCodes,
  services = c(
    "overview",
    "categorization"),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(rppoCodes)
  assert_vector(services)
  assert_string(apiKey)
  
  .checkServices(services, eval(formals()$services))
  
  rppoData_ <- map(rppoCodes, ~ {
    rppoCode_ <- .x
    
    map(services, ~ .fetchService(
      basePath = "/rppo",
      apiKey = apiKey,
      code = rppoCode_,
      service = .x)) |>
      set_names(services)}) |>
    
    set_names(rppoCodes)
  
  rppoData_ <- .mergeBatch(
    datasets = rppoData_,
    parentColumnName = "queried_rppo_code")
  
  return(rppoData_)
}

#' Query the EPPO API Tools endpoints.
#' 
#' This function queries the Tools endpoints of the EPPO Global Database via
#' REST API. The function sequentially queries all specified `services` and
#' returns the extracted data through a list of dataframes.
#'
#' @param services `character` (vector). One or more Tools services to query.
#'   Common service names include: `name2codes`. A validation step ensures that
#'   all provided services match the supported service names.
#' 
#'   By default: all services.
#' 
#' @param params `list`. A named list of query parameters to include in the
#'   request. The list of available parameters can be accessed via the
#'   [EPPO API Documentation](https://data.eppo.int/ui/#/docs/GDAPI)
#'   platform.
#' 
#'   By default: no parameters, you need to specify the required ones.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content.
#' 
#' @importFrom checkmate assert_vector assert_list assert_string assert
#' @importFrom purrr map2 set_names
#'
#' @examples
#' \dontrun{
#' # Get the EPPO codes associated to the name Bemisia tabaci.
#' toolsData_ <- tools(params = list(
#'   name2codes = list(
#'     onlyPreferred = FALSE,
#'     name = "Bemisia tabaci")))
#' }
#' 
#' @export
#' 
tools <- function(
  services = c(
    "name2codes"),
  params = list(
    "name2codes" = list()),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(services)
  assert_list(params)
  assert_string(apiKey)
  assert(length(services) == length(params))
  
  .checkServices(services, eval(formals()$services))
  
  toolsData_ <- map2(services, params[services], ~ .fetchService(
    basePath = "/tools",
    apiKey = apiKey,
    service = .x,
    params = .y)) |>
    set_names(services)
  
  toolsData_ <- .valuesToTibbles(values = toolsData_)
  
  return(toolsData_)
}

#' Query the EPPO API Taxons endpoints.
#' 
#' This function queries the Taxons endpoints of the EPPO Global Database via
#' REST API. The function sequentially queries all specified `services` and
#' returns the extracted data through a list of dataframes.
#'
#' @param services `character` (vector). One or more Taxons services to query.
#'   Common service names include: `list`, `reporting` and `article`. A
#'   validation step ensures that all provided services match the supported
#'   service names.
#' 
#'   By default: all services.
#' 
#' @param params `list`. A named list of query parameters to include in the
#'   request. The list of available parameters can be accessed via the
#'   [EPPO API Documentation](https://data.eppo.int/ui/#/docs/GDAPI)
#'   platform.
#' 
#'   By default: no parameters, you need to specify the required ones.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content.
#' 
#' @importFrom checkmate assert_vector assert_list assert_string assert
#' @importFrom purrr map2 set_names
#'
#' @examples
#' \dontrun{
#' # Get the list of reporting service issues.
#' reportingServiceData_ <- reportingService(
#'   services = c("list"),
#'   params = list(
#'     "list" = list()))
#'
#' # Get a specific reporting service issue by ID.
#' reportingServiceData_ <- reportingService(
#'   services = c("reporting"),
#'   params = list(
#'     reporting = list(
#'       reporting_id = 10)))
#'
#' # Get a specific article by ID.
#' reportingServiceData_ <- reportingService(
#'   services = c("article"),
#'   params = list(
#'     article = list(
#'       article_id = 234)))
#'
#' # Get a specific reporting service issue and a specific article by ID.
#' reportingServiceData_ <- reportingService(
#'   services = c("reporting", "article"),
#'   params = list(
#'     reporting = list(reporting_id = 10),
#'     article = list(article_id = 234)
#'   ))
#' }
#' 
#' @export
#' 
reportingService <- function(
  services = c(
    "list",
    "reporting",
    "article"),
  params = list(
    "list" = list(),
    "reporting" = list(),
    "article" = list()),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(services)
  assert_list(params)
  assert_string(apiKey)
  assert(length(services) == length(params))
  
  .checkServices(services, eval(formals()$services))
  
  reportingServiceData_ <- map2(services, params[services], ~ .fetchService(
    basePath = "/reportings",
    apiKey = apiKey,
    service = .buildReportingServicePath(
      service = .x,
      params = .y))) |>
    set_names(services)
  
  reportingServiceData_ <- .valuesToTibbles(values = reportingServiceData_)
  
  return(reportingServiceData_)
}

test_that("The services must be specified in a vector", {
  expect_error(
    reportingService(
      services = 123))
})

test_that("The parameters must be specified in a list", {
  expect_error(
    reportingService(
      params = 123))
})

test_that("The API key must be a string", {
  expect_error(
    reportingService(
      apiKey = 123))
})

test_that("The number of parameters must be equal to the number of services", {
  expect_error(
    reportingService(
      services = c("list", "article"),
      params = list()))
})

test_that("The result must be a list", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_true(
        is.list(
          reportingService(
            services = c("list"),
            params = list("list" = list()))))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("The result must be a list", {
  skip_on_cran()
  
  expect_true(
    is.list(
      reportingService(
        services = c("list"),
        params = list("list" = list()))))
})

test_that("The names in the list are equal to the specified services", {
  services_ <- c(
    "list")
  
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_equal(
        names(
          reportingService(
            services = services_,
            params = list("list" = list()))),
        services_)
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("The names in the list are equal to the specified services", {
  skip_on_cran()
  
  services_ <- c(
    "list")
  
  expect_equal(
    names(
      reportingService(
        services = services_,
        params = list("list" = list()))),
    services_)
})

test_that("Expect no errors if correct parameters are specified", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_no_error(
        reportingService(
          services = c("reporting"),
          params = list(
            "reporting" = list(
              reporting_id = 10))))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect no errors if correct parameters are specified", {
  skip_on_cran()
  
  expect_no_error(
    reportingService(
      services = c("reporting"),
      params = list(
        "reporting" = list(
          reporting_id = 10))))
})

test_that("If a bad service is provided than an error is thrown", {
  expect_error(
    reportingService(
      services = c("badService"),
      params = list("badService" = list())))
})

test_that("If a bad API key is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (apiKey == "BAD_API_KEY") { stop() }
    }, {
      expect_error(
        reportingService(
          services = c("list"),
          params = list("list" = list()),
          apiKey = "BAD_API_KEY"))
    }
  )
})

# This test performs real requests to the EPPO API.
test_that("If a bad API key is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    reportingService(
      services = c("list"),
      params = list("list" = list()),
      apiKey = "BAD_API_KEY"))
})

test_that("If a bad environment variable is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (apiKey == '') { stop() }
    }, {
      expect_error(
        reportingService(
          services = c("list"),
          params = list("list" = list()),
          apiKey = Sys.getenv("BAD_ENV_VAR")))
    }
  )
})

# This test performs real requests to the EPPO API.
test_that("If a bad environment variable is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    reportingService(
      services = c("list"),
      params = list("list" = list()),
      apiKey = Sys.getenv("BAD_ENV_VAR")))
})

#' Reshape countries–states reference data.
#'
#' This internal utility restructures the `countriesStates` output returned by
#' the EPPO References API. The function removes metadata from the raw API
#' response, converts each country element into a tibble, adds the corresponding
#' ISO country code, and finally binds all results into a single dataframe.
#'
#' @param countriesStatesData `list`. A list resulting from a `countriesStates`
#'   API query.
#'
#' @return A single tibble combining all countries' states. The returned tibble
#'   includes a new column `country_iso` indicating the ISO code for each group
#'   of rows.
#' 
#' @importFrom checkmate assert_list
#' @importFrom dplyr mutate relocate bind_rows
#' @importFrom purrr imap
#'
#' @keywords internal
#' @noRd
#'
.reshapeCountriesStates <- function(countriesStatesData) {
  
  assert_list(countriesStatesData)
  
  queriedOn_ <- countriesStatesData$queried_on
  queriedURL_ <- countriesStatesData$queried_url
  
  countriesStatesData$queried_url <- NULL
  countriesStatesData$queried_on <- NULL
  
  countriesStatesData <- .valuesToTibbles(values = countriesStatesData)
  
  countriesStatesData <- imap(countriesStatesData, ~ .x |>
   mutate(country_iso = .y) |>
   relocate(country_iso, .before = 1))
  
  countriesStatesData <- bind_rows(countriesStatesData)
  
  countriesStatesData$queried_on <- queriedOn_
  countriesStatesData$queried_url <- queriedURL_
  
  return(countriesStatesData)
}

#' Build an EPPO API endpoint path.
#'
#' This helper function constructs an endpoint path for retrieving data from the
#' EPPO API. The result must be appended to the EPPO API base URL. It allows for
#' the optional inclusion of a specific code and/or service name, depending on
#' the desired API resource. The function is based on the fact that EPPO API
#' endpoints follow the pattern:
#' `{base path}/{resource identifier}/{service}`.
#'
#' @param basePath `character` (string). The base path, starting with "/" 
#'   (e.g., `"/taxons/taxon"`). Must match the pattern `"^/"`.
#' 
#' @param code `character` (string, optional). The resource identifier (e.g. an
#'   EPPO code or an ISO code). If provided, it will be appended to the base
#'   path.
#' 
#' @param service `character` (string, optional). The desired API service. If
#'   provided, it will be appended to the path after the resource identifier
#'   (if any, otherwise after the base path).
#'
#' @return A string representing the complete endpoint path to be used in an API
#'   request.
#' 
#' @importFrom checkmate assert_string
#' @importFrom glue glue
#'
#' @keywords internal
#' @noRd
#' 
.buildEndpoint <- function(basePath, code = NULL, service = NULL) {

  assert_string(basePath, pattern = "^/")
  if (!is.null(code)) assert_string(code)
  if (!is.null(service)) assert_string(service)
  
  endpointParts_ <- c(basePath, code, service)
  
  endpointPath_ <- paste(endpointParts_, collapse = '/')
  endpointPath_ <- gsub("//", '/', endpointPath_)
  
  return(endpointPath_)
}

test_that("The base path must be a string", {
  expect_error(
    .buildEndpoint(
      basePath = 123))
})

test_that("The base path must start with a slash", {
  expect_error(
    .buildEndpoint(
      basePath = "taxons/taxon"))
})

test_that("The resource identifier must be a string", {
  expect_error(
    .buildEndpoint(
      basePath = "/taxons/taxon",
      code = 123))
})

test_that("The service must be a string", {
  expect_error(
    .buildEndpoint(
      basePath = "/taxons/taxon",
      code = "BEMITA",
      service = 123))
})

test_that("If only the base path is given, the result must be the base path", {
  expect_equal(
    .buildEndpoint(
      basePath = "/taxons/taxon"),
    "/taxons/taxon")
})

test_that("If a resource identifier is given, the result must include it", {
  expect_equal(
    .buildEndpoint(
      basePath = "/taxons/taxon",
      code = "BEMITA"),
    "/taxons/taxon/BEMITA")
})

test_that("If a service is given, the result must include it", {
  expect_equal(
    .buildEndpoint(
      basePath = "/country",
      code = "FR",
      service = "overview"),
    "/country/FR/overview")
})

test_that("If no resource identifier is specified, it must be avoided", {
  expect_equal(
    .buildEndpoint(
      basePath = "/reportings",
      service = "list"),
    "/reportings/list")
})

#' Fetch data from a specific EPPO API service.
#'
#' This function retrieves data from a specific service of the EPPO API. It
#' builds the appropriate endpoint according to the EPPO API structure, then
#' queries the API.
#'
#' @param basePath `character` (string). The base path, starting with "/" 
#'   (e.g., `"/taxons/taxon"`). Must match the pattern `"^/"`.
#' 
#' @param apiKey `character` (string) The API key used for authentication.
#' 
#' @param code `character` (string, optional). The resource identifier (e.g. an
#'   EPPO code or an ISO code). If provided, it will be appended to the base
#'   path.
#' 
#' @param service `character` (string, optional). The desired API service. If
#'   provided and not empty, it will be appended to the path after the resource
#'   identifier (if any, otherwise after the base path).
#' 
#' @param params `list` (optional). A named list of query parameters to include
#'   in the request.
#'
#' @return A structure containing the data returned by the specified EPPO API
#'   service.
#' 
#' @importFrom checkmate assert_string assert_list
#'
#' @keywords internal
#' @noRd
#' 
.fetchService <- function(
  basePath,
  apiKey,
  code = NULL,
  service = NULL,
  params = list()) {
  
  assert_string(basePath, pattern = "^/")
  assert_string(apiKey)
  if (!is.null(code)) assert_string(code)
  if (!is.null(service)) assert_string(service)
  assert_list(params)
  
  serviceEndpoint_ <- .buildEndpoint(
    basePath = basePath,
    code = code,
    service = service)
  
  serviceData_ <- .query(
    endpoint = serviceEndpoint_,
    apiKey = apiKey,
    params = params)
  
  return(serviceData_)
}

test_that("The base path must be a string", {
  expect_error(
    .fetchService(
      basePath = 123))
})

test_that("The base path must start with a slash", {
  expect_error(
    .fetchService(
      basePath = "taxons/taxon"))
})

test_that("The resource identifier must be a string", {
  expect_error(
    .fetchService(
      basePath = "/taxons/taxon",
      code = 123))
})

test_that("The service must be a string", {
  expect_error(
    .fetchService(
      basePath = "/taxons/taxon",
      code = "BEMITA",
      service = 123))
})

test_that("The parameters must be specified in a list", {
  expect_error(
    .fetchService(
      basePath = "/taxons/taxon",
      code = "BEMITA",
      service = "names",
      params = 123))
})

test_that("The API key must be a string", {
  expect_error(
    .fetchService(
      basePath = "/taxons/taxon",
      apiKey = 123))
})

test_that("Expect no errors if the parameters are correct", {
  with_mocked_bindings(
    .query = function(endpoint, apiKey, params) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_no_error(
        .fetchService(
          basePath = "/taxons/taxon",
          apiKey = Sys.getenv("EPPO_API_KEY"),
          code = "BEMITA",
          service = "overview"))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect no errors if the parameters are correct", {
  skip_on_cran()
  
  expect_no_error(
    .fetchService(
      basePath = "/taxons/taxon",
      apiKey = Sys.getenv("EPPO_API_KEY"),
      code = "BEMITA",
      service = "overview"))
})

#' Query the EPPO API General endpoint.
#' 
#' This function queries the General endpoints of the EPPO Global Database via
#' REST API. The function sequentially queries all specified `services` and
#' returns the extracted data through a list of dataframes.
#'
#' @param services `character` (vector). One or more General services to query.
#'   Common service names include: `status`. A validation step ensures that all
#'   provided services match the supported service names.
#' 
#'   By default: all services.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content.
#' 
#' @importFrom checkmate assert_vector assert_string
#' @importFrom purrr map set_names
#'
#' @examples
#' \dontrun{
#' # Get information about system health status.
#' generalData_ <- general()
#' }
#' 
#' @export
#' 
general <- function(
  services = c("status"),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(services)
  assert_string(apiKey)
  
  .checkServices(services, eval(formals()$services))
  
  generalData_ <- map(services, ~ .fetchService(
    basePath = "/",
    apiKey = apiKey,
    service = .x)) |>
    set_names(services)
  
  generalData_ <- .valuesToTibbles(values = generalData_)
  
  return(generalData_)
}

test_that("The services must be specified in a vector", {
  expect_error(
    general(
      services = 123))
})

test_that("The API key must be a string", {
  expect_error(
    general(
      apiKey = 123))
})

test_that("The result must be a list", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_true(
        is.list(
          general()))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("The result must be a list", {
  skip_on_cran()
  
  expect_true(
    is.list(
      general()))
})

test_that("The names in the list are equal to the specified services", {
  services_ <- c(
    "status")
  
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_equal(
        names(
          general()),
        services_)
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("The names in the list are equal to the specified services", {
  skip_on_cran()
  
  services_ <- c(
    "status")
  
  expect_equal(
    names(
      general()),
    services_)
})

test_that("If a bad service is provided than an error is thrown", {
  expect_error(
    general(
      services = c("badService")))
})

test_that("If a bad API key is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (apiKey == "BAD_API_KEY") { stop() }
    }, {
      expect_error(
        general(
          apiKey = "BAD_API_KEY"))
    }
  )
})

# This test performs real requests to the EPPO API.
test_that("If a bad API key is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    general(
      apiKey = "BAD_API_KEY"))
})

test_that("If a bad environment variable is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (apiKey == '') { stop() }
    }, {
      expect_error(
        general(
          apiKey = Sys.getenv("BAD_ENV_VAR")))
    }
  )
})

# This test performs real requests to the EPPO API.
test_that("If a bad environment variable is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    general(
      apiKey = Sys.getenv("BAD_ENV_VAR")))
})

#' Create a complete and uniform taxonomy dataframe.
#'
#' This function normalizes the taxonomy returned by the EPPO service,
#' producing a uniform structure that includes all possible taxonomic
#' categories, even when some of them are not present in the original result.
#'
#' The function uses an internal set of known taxonomic ranks and performs
#' a left join with the `taxonomy` dataframe, filling missing ranks with `NA`.
#' After the join, the column `level` from the original taxonomy is removed,
#' since the rank order is fully defined by the internal reference list. The
#' column `queried_eppo_code` NA values are filled with the right code.
#'
#' @param taxonomyData `tibble`. A dataframe containing taxonomy data provided
#'   by the EPPO service for a given EPPO code.
#'
#' @return A dataframe where each row represents one of the expected taxonomic
#'   ranks. Fields corresponding to ranks not present in the original taxonomy
#'   are filled with `NA`. The `level` column is excluded from the output. The
#'   column `queried_eppo_code` NA values are filled with the right code.
#'
#' @importFrom dplyr left_join select any_of
#' @importFrom checkmate assert_data_frame assert_names assert_true
#' @importFrom tibble tibble
#'
#' @examples
#' \dontrun{
#' # Retrieve taxonomy data from the EPPO service.
#' taxonData_ <- taxon(eppoCodes = c("BEMITA"), services = c("taxonomy"))
#'
#' # Create a uniform taxonomy with all ranks.
#' taxonomyData_ <- uniformTaxonomy(taxonomyData = taxonData_$taxonomy)
#' }
#' 
#' @export
#' 
uniformTaxonomy <- function(taxonomyData) {
  
  assert_data_frame(taxonomyData)
  assert_names(names(taxonomyData), must.include = c(
    "queried_eppo_code", "type"))
  assert_true(any(!is.na(taxonomyData$queried_eppo_code)))
  
  taxonomyTypes_ <- tibble(
    type = c(
      "Kingdom",
      "Phylum",
      "Subphylum",
      "Class",
      "Subclass",
      "Order",
      "Suborder",
      "Family",
      "Subfamily",
      "Genus",
      "Species"
    )
  )
  
  taxonomyData <- taxonomyTypes_ |>
    left_join(taxonomyData, by = "type") |>
    select(-any_of("level"))
  
  queried_eppo_code_ <- taxonomyData$queried_eppo_code[
    !is.na(taxonomyData$queried_eppo_code)][1]
  
  taxonomyData <- taxonomyData |> mutate(
    queried_eppo_code = {
      ifelse(
        is.na(taxonomyData$queried_eppo_code),
        queried_eppo_code_,
        taxonomyData$queried_eppo_code)
    }
  )
  
  return(taxonomyData)
}

test_that("The taxonomy must be provided in a dataframe", {
  expect_error(
    uniformTaxonomy(
      taxonomyData = 123))
})

test_that("The taxonomy dataframe must contain mandatory columns", {
  expect_error(
    uniformTaxonomy(
      taxonomyData = tibble()))
})

test_that("The queried_eppo_code column must have at least a non-NA value", {
  expect_error(
    uniformTaxonomy(
      taxonomyData = tibble(queried_eppo_code = c(NA, NA, NA))))
})

test_that("The output must be a dataframe", {
  taxonomyData_ <- tibble(
    queried_eppo_code = c("A", "B", "C"),
    eppocode = c("D", "E", "F"),
    prefname = c("G", "H", "I"),
    level = c(1, 2, 3),
    type = c("J", "K", "L"),
    queried_url = c("M", "N", "O"),
    queried_on = c("P", "Q", "R")
  )
  
  expect_true(
    is.data.frame(
      uniformTaxonomy(
        taxonomyData = taxonomyData_)))
})

#' Merge nested lists of dataframes into tibbles.
#'
#' This function takes a named list of named lists containing data frames or
#' lists of values, converts each inner element into a tibble, adds a column
#' with the name of the parent list, and combines the inner elements across all
#' parent lists into unified tibbles for each child name.
#'
#' @param datasets `list`. A named list of named lists. Each inner list should
#'   contain dataframes or objects convertible to tibbles.
#' 
#' @param parentColumnName `character` (string). A string specifying the name of
#'   the new column that will contain the name of the parent list.
#'
#' @return A named list of tibbles. Each tibble corresponds to a unique child
#'   name across the nested lists, with an additional column containing the
#'   parent list names.
#' 
#' @importFrom checkmate assert_list assert_string
#' 
#' @keywords internal
#' @noRd
#'
.mergeBatch <- function(datasets, parentColumnName) {
  
  assert_list(datasets)
  assert_string(parentColumnName)
  
  datasets <- .valuesToTibblesBatch(
    nestedValues = datasets,
    parentColumnName = parentColumnName)
  
  groupedDatasets <- .groupByServices(datasets = datasets)
  
  return(groupedDatasets)
}

#' Query the EPPO API Taxon endpoints.
#' 
#' This function queries the Taxon endpoints of the EPPO Global Database via
#' REST API for one or more EPPO code(s) and one or more service(s). For each
#' EPPO code in `eppoCodes`, the function sequentially queries all specified
#' `services` and returns the extracted data through a list of dataframes.
#' 
#' @param eppoCodes `character` (vector). One or more EPPO codes to query.
#'
#'   EPPO codes are standardized identifiers for plants and pests (including
#'   pathogens). These alphanumeric codes, typically of 5 to 6 letters long,
#'   serve as unique, mnemonic abbreviations of scientific names. The EPPO
#'   coding system facilitates the management and exchange of organism names
#'   across databases and IT systems. EPPO codes and their corresponding data
#'   can be accessed via the [EPPO Data Services](https://data.eppo.int/)
#'   platform.
#'
#' @param services `character` (vector). One or more Taxon services to query.
#'   Common service names include: `overview`, `infos`, `names`, `taxonomy`,
#'   `categorization`, `kingdom`, `hosts`, `pests`, `vectors`, `vectorof`,
#'   `bca`, `bcaof`, `photos`, `reporting_articles`, `documents`, `standards`
#'   and `distribution`. A validation step ensures that all provided services
#'   match the supported service names.
#' 
#'   By default: all services.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content for all the specified EPPO
#'   codes.
#' 
#' @importFrom checkmate assert_vector assert_string
#' @importFrom purrr map set_names
#'
#' @examples
#' \dontrun{
#' # Get all information about Bemisia tabaci.
#' taxonData_ <- taxon(eppoCodes = c("BEMITA"))
#' 
#' # Get names data about Bemisia tabaci.
#' taxonData_ <- taxon(eppoCodes = c("BEMITA"), services = c("names"))
#' 
#' # Get taxonomy and categorization data about Bemisia tabaci and Gossypium
#' # hirsutum.
#' taxonData_ <- taxon(
#'   eppoCodes = c("BEMITA", "GOSHI"),
#'   services = c("taxonomy", "categorization"))
#' }
#' 
#' @export
#' 
taxon <- function(
  eppoCodes,
  services = c(
    "overview",
    "infos",
    "names",
    "taxonomy",
    "categorization",
    "kingdom",
    "hosts",
    "pests",
    "vectors",
    "vectorof",
    "bca",
    "bcaof",
    "photos",
    "reporting_articles",
    "documents",
    "standards",
    "distribution"),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(eppoCodes)
  assert_vector(services)
  assert_string(apiKey)
  
  .checkServices(services, eval(formals()$services))
  
  taxonData_ <- map(eppoCodes, ~ {
    eppoCode_ <- .x
    
    map(services, ~ .fetchService(
      basePath = "/taxons/taxon",
      apiKey = apiKey,
      code = eppoCode_,
      service = .x)) |>
      set_names(services)}) |>
    
    set_names(eppoCodes)
  
  taxonData_ <- .mergeBatch(
    datasets = taxonData_,
    parentColumnName = "queried_eppo_code")
  
  return(taxonData_)
}

#' Rebuild EPPO Taxons response data.
#'
#' This helper function processes the raw response from the EPPO Taxons services 
#' and restructures it into a format suitable for conversion into tibbles.
#' It extracts metadata and flattens the nested data structure.
#'
#' @param taxonsData `list`. The raw output from the `taxons()` function.
#'
#' @return A restructured list in which the elements has been converted into
#'   tibbles, with additional columns for `queried_on` and `queried_url` to
#'   preserve metadata from the original response.
#'
#' @importFrom checkmate assert_list
#' @importFrom dplyr mutate
#'
#' @details
#' Since only the `list` service is currently supported, this function focuses 
#' exclusively on that element. It extracts the `queried_on` and `queried_url` 
#' fields from the raw response, replaces the `data` element with the actual 
#' content, and then converts it into a tibble using `.valuesToTibbles()`.
#'
#' @keywords internal
#' @noRd
#' 
.rebuildTaxons <- function(taxonsData) {
  
  assert_list(taxonsData)
  
  queriedOn_ <- taxonsData$list$queried_on
  queriedURL_ <- taxonsData$list$queried_url
  taxonsData$list <- taxonsData$list$data
  
  taxonsData <- .valuesToTibbles(values = taxonsData)
  
  taxonsData$list <- taxonsData$list |> mutate(
    queried_on = queriedOn_,
    queried_url = queriedURL_
  )
  
  return(taxonsData)
}

test_that("The data parameter must be a list", {
  expect_error(
    .rebuildTaxons(
      taxonsData = 123))
})

test_that("The result must be a list", {
  expect_true(
    is.list(
      .rebuildTaxons(
        taxonsData = list(
          "list" = list(
            data = list(x = 1:3, y = 4:6),
            queried_url = "https://example.org",
            queried_on = Sys.time())))))
})

#' Enrich API response data.
#'
#' This helper function takes a list containing API response data and enriches
#' it with metadata, including the URL queried and the timestamp of the query.
#'
#' @param responseData `list` or `data.frame`. A structure representing the API
#'   response.
#'
#' @param url `character` (string). The URL that was queried.
#'
#' @return A list containing the original API response data, along with
#'   two additional fields: `queried_on`, the timestamp when the data was
#'   queried, and `queried_url`, the URL used for the API request.
#' 
#' @importFrom checkmate assert assert_string check_list check_data_frame
#' 
#' @keywords internal
#' @noRd
#'
.enrichResponse <- function(responseData, url) {
  
  assert(check_list(responseData), check_data_frame(responseData))
  assert_string(url)
  
  responseData$queried_on <- Sys.time()
  responseData$queried_url <- url
  
  return(responseData)
}

test_that("The response data must be a list or a dataframe", {
  expect_error(
    .enrichResponse(
      responseData = 123,
      url = ""))
})

test_that("The URL must be a string", {
  expect_error(
    .enrichResponse(
      responseData = list(),
      url = 123))
})

test_that("Expect no errors if the parameters are correct", {
  enrichedData_ <- .enrichResponse(
    responseData = list(),
    url = "https://example.org")
  expect_true("queried_on" %in% names(enrichedData_))
  expect_equal(enrichedData_$queried_url, "https://example.org")
})

#' Parse a JSON API response.
#' 
#' This helper function parses the JSON body of an API response.
#' 
#' @param response `httr2_response`. An HTTP response object returned by
#'   `httr2::req_perform()`.
#' 
#' @return A data structure containing the parsed JSON response.
#' 
#' @details
#' If the API response body cannot be parsed as valid JSON, the function raises
#' an error, including the original parsing error message.
#' 
#' @importFrom checkmate assert_class
#' @importFrom jsonlite fromJSON
#' @importFrom httr2 resp_body_string
#' @importFrom cli cli_abort
#' 
#' @keywords internal
#' @noRd
#' 
.parseResponse <- function(response) {
  
  assert_class(response, "httr2_response")
  
  responseData_ <- tryCatch(
    fromJSON(resp_body_string(response), flatten = TRUE),
    error = function(e) {
      cli_abort(c(
        "Failed to parse API response.",
        "x" = conditionMessage(e)
      ))
    }
  )
  
  return(responseData_)
}

test_that("The response parameter must be an httr2_response object", {
  expect_error(
    .parseResponse(
      response = 123))
})

test_that("Expect an error if the response delivers invalid data", {
  response_ <- httr2::response(
    status_code = 200,
    url = "https://api.eppo.int/gd/v2/status",
    method = "GET",
    headers = list("Content-Type" = "application/json"),
    body = charToRaw("{'a': 1,"))
  
  expect_error(
    .parseResponse(
      response = response_))
})

test_that("Expect no error if all the parameters are correct", {
  response_ <- httr2::response(
    status_code = 200,
    url = "https://api.eppo.int/gd/v2/status",
    method = "GET",
    headers = list("Content-Type" = "application/json"),
    body = charToRaw("{\"a\": 1}"))
  
  expect_no_error(
    .parseResponse(
      response = response_))
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect no error if all the parameters are correct", {
  skip_on_cran()
  
  request_ <- .buildRequest(
    url = "https://api.eppo.int/gd/v2/taxons/taxon/BEMITA/overview",
    apiKey = Sys.getenv("EPPO_API_KEY"))
  
  response_ <- .performRequest(request = request_)
  
  expect_no_error(
    .parseResponse(
      response = response_))
})

#' Rebuild EPPO References response data.
#'
#' This internal helper function restructures the raw output returned by the
#' EPPO References API for each queried service. Depending on the service name,
#' the function applies the appropriate post-processing function to convert the
#' API response into clean tibbles.
#'
#' @param referencesData `list`. A named list containing the raw results
#'   returned by the API for each requested service. Each list element
#'   corresponds to a service and contains the unprocessed result of
#'   `.fetchService()`.
#'
#' @return A list where each element is a cleaned and flattened tibble
#'  representing the retrieved data for the corresponding service.
#' 
#' @importFrom checkmate assert_list
#' @importFrom purrr imap
#'
#' @details
#' The purpose of this function is to harmonize the structure of the various
#' Reference endpoints, ensuring that the final output from `references()` is
#' consistently formatted and ready for use.
#'
#' @keywords internal
#' @noRd
#' 
.rebuildReferences <- function(referencesData) {
  
  assert_list(referencesData)
  
  referencesData <- imap(referencesData,
    ~ if (.y == "countriesStates") {
      .reshapeCountriesStates(countriesStatesData = .x)
    } else {
      .valueToTibble(value = .x)
    })
  
  return(referencesData)
}

#' Build an HTTP GET request for the EPPO API.
#' 
#' This helper function constructs and configures an HTTP GET request object
#' using the `httr2` package. It sets the required headers, authentication key, 
#' and custom error-handling behavior based on allowed HTTP status codes.
#' 
#' @param url `character` (string). The full API endpoint URL to query.
#' 
#' @param apiKey `character` (string) The API key used for authentication, sent
#'   via the `X-Api-Key` header.
#' 
#' @param doNotFailOn `numeric` (vector, optional). The HTTP status codes that
#'   should not trigger an error.
#' 
#'   By default: only status code 200.
#' 
#' @param params `list` (optional). A named list of query parameters to include
#'   in the request.
#' 
#' @return An `httr2_request` object representing the configured API request.
#' 
#' @details
#' The function sets the HTTP method to `GET` and includes the following
#' headers:
#' \itemize{
#'  \item `X-Api-Key` - the API Key used for authentication.
#'  \item `Accept` - set to `application/json` to request JSON responses.
#' }
#' 
#' A custom error handler is set via `req_error()`, ensuring that only responses
#' with status codes *not* included in `doNotFailOn` are treated as errors.
#' 
#' @importFrom checkmate assert_string assert_vector assert_list
#' @importFrom httr2 request req_method req_headers req_error resp_status
#' req_url_query
#' 
#' @keywords internal
#' @noRd
#' 
.buildRequest <- function(url, apiKey, doNotFailOn = c(200), params = list()) {
  
  assert_string(url)
  assert_string(apiKey)
  assert_vector(doNotFailOn)
  assert_list(params)
  
  request_ <- request(url) |>
    req_method("GET") |>
    req_headers(
      "X-Api-Key" = apiKey,
      "Accept" = "application/json"
    ) |>
    req_url_query(!!!params) |>
    req_error(is_error = function(response_) {
      return(!(resp_status(response_) %in% doNotFailOn))
    })
  
  return(request_)
}

test_that("The URL must be a string", {
  expect_error(
    .buildRequest(
      url = 123,
      apiKey = "EPPO_API_KEY"))
})

test_that("The parameters must be specified in a list", {
  expect_error(
    .buildRequest(
      url = "https://example.org",
      apiKey = "EPPO_API_KEY",
      params = 123))
})

test_that("The API key must be a string", {
  expect_error(
    .buildRequest(
      url = "https://example.org",
      apiKey = 123))
})

test_that("The list of non-failure codes must be a vector", {
  expect_error(
    .buildRequest(
      url = "https://example.org",
      apiKey = "EPPO_API_KEY",
      doNotFailOn = NULL))
})

test_that("The result must be a httr2_request", {
  expect_s3_class(
    .buildRequest(
      url = "https://example.org",
      apiKey = "EPPO_API_KEY"),
    "httr2_request")
})

test_that("The request parameter must be an httr2_request object", {
  expect_error(
    .performRequest(
      request = 123))
})

test_that("Expect an error if the request is malformed", {
  request_ <- .buildRequest(
    url = "https://invalid_domain",
    apiKey = "api_key")
  
  expect_error(
    .performRequest(
      request = request_))
})

test_that("Expect no error if all the request object is correct", {
  request_ <- .buildRequest(
    url = "https://api.eppo.int/gd/v2/taxons/taxon/BEMITA/overview",
    apiKey = "EPPO_API_KEY")
  
  response_ <- httr2::response(
    status_code = 200)
  
  with_mocked_bindings(
    req_perform = function(request) {
      return(response_)
    }, {
      expect_no_error(
        .performRequest(
          request = request_))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect no error if all the request object is correct", {
  skip_on_cran()
  
  request_ <- .buildRequest(
    url = "https://api.eppo.int/gd/v2/taxons/taxon/BEMITA/overview",
    apiKey = Sys.getenv("EPPO_API_KEY"))
  
  expect_no_error(
    .performRequest(
      request = request_))
})

test_that("The function must behave as expected for error status codes", {
  request_ <- .buildRequest(
    url = "https://api.eppo.int/gd/v2/status",
    apiKey = "EPPO_API_KEY")
  
  response_ <- httr2::response(
    status_code = 403)
  
  is_error_fn_ <- request_$policies$error_is_error
  
  expect_true(is_error_fn_(response_))
})

test_that("The result must be an httr2_response object", {
  request_ <- .buildRequest(
    url = "https://api.eppo.int/gd/v2/taxons/taxon/BEMITA/overview",
    apiKey = "EPPO_API_KEY")
  
  response_ <- httr2::response(
    status_code = 200)
  
  with_mocked_bindings(
    req_perform = function(request) {
      return(response_)
    }, {
      expect_s3_class(
        .performRequest(
          request = request_),
        "httr2_response")
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("The result must be an httr2_response object", {
  skip_on_cran()
  
  request_ <- .buildRequest(
    url = "https://api.eppo.int/gd/v2/taxons/taxon/BEMITA/overview",
    apiKey = Sys.getenv("EPPO_API_KEY"))
  
  expect_s3_class(
    .performRequest(
      request = request_),
    "httr2_response")
})

#' Query the EPPO API Country endpoints.
#' 
#' This function queries the Country endpoints of the EPPO Global Database via
#' REST API for one or more ISO code(s) and one or more service(s). For each
#' ISO code in `isoCodes`, the function sequentially queries all specified
#' `services` and returns the extracted data through a list of dataframes.
#' 
#' @param isoCodes `character` (vector). One or more ISO codes to query.
#' 
#'   ISO country codes are standardized identifiers for countries and
#'   territories. These codes, typically consisting of two-letter, provide
#'   unique and consistent abbreviations of country names. The ISO coding system
#'   enables easy management, exchange, and integration of country information
#'   across databases and IT systems. ISO country codes and their related data
#'   can be accessed via the
#'   [ISO Online Browsing Platform](https://www.iso.org/obp/ui/#search).
#'
#' @param services `character` (vector). One or more Country services to query.
#'   Common service names include: `overview`, `categorization` and `presence`.
#'   A validation step ensures that all provided services match the supported
#'   service names.
#' 
#'   By default: all services.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content for all the specified ISO
#'   codes.
#' 
#' @importFrom checkmate assert_vector assert_string
#' @importFrom purrr map set_names
#'
#' @examples
#' \dontrun{
#' # Get all information about France.
#' countryData_ <- country(isoCodes = c("FR"))
#' 
#' # Get categorization data about France.
#' countryData_ <- country(isoCodes = c("FR"), services = c("categorization"))
#' 
#' # Get basic data about France and Italy.
#' countryData_ <- country(isoCodes = c("FR", "IT"), services = c("overview"))
#' }
#' 
#' @export
#' 
country <- function(
  isoCodes,
  services = c(
    "overview",
    "categorization",
    "presence"),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(isoCodes)
  assert_vector(services)
  assert_string(apiKey)
  
  .checkServices(services, eval(formals()$services))
  
  countryData_ <- map(isoCodes, ~ {
    isoCode_ <- .x
    
    map(services, ~ .fetchService(
      basePath = "/country",
      apiKey = apiKey,
      code = isoCode_,
      service = .x)) |>
      set_names(services)}) |>
    
    set_names(isoCodes)
  
  countryData_ <- .mergeBatch(
    datasets = countryData_,
    parentColumnName = "queried_iso_code")
  
  return(countryData_)
}

test_that("The ISO codes must be specified in a vector", {
  expect_error(
    country(
      isoCodes = 123))
})

test_that("The services must be specified in a vector", {
  expect_error(
    country(
      isoCodes = c("FR"),
      services = 123))
})

test_that("The API key must be a string", {
  expect_error(
    country(
      isoCodes = c("FR"),
      apiKey = 123))
})

test_that("If parameters are correct, a list is provided", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      expect_true(
        is.list(
          country(
            isoCodes = c("FR"))))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("If parameters are correct, a list is provided", {
  skip_on_cran()
  
  expect_true(
    is.list(
      country(
        isoCodes = c("FR"))))
})

test_that("The names in the list are equal to the specified services", {
  services_ <- c(
    "overview",
    "categorization")
  
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      return(list(queried_on = "", queried_url = ""))
    }, {
      data_ <- country(
        isoCodes = c("FR"),
        services = services_)
      
      expect_equal(
        names(data_),
        services_)
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("The names in the list are equal to the specified services", {
  skip_on_cran()
  
  services_ <- c(
    "overview",
    "categorization")
  
  data_ <- country(
    isoCodes = c("FR"),
    services = services_)
  
  expect_equal(
    names(data_),
    services_)
})

test_that("If a bad ISO code is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (code == "BAD_ISO_CODE") { stop() }
    }, {
      expect_error(
        country(
          isoCodes = c("BAD_ISO_CODE")))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("If a bad ISO code is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    country(
      isoCodes = c("BAD_ISO_CODE")))
})

test_that("If a bad service is provided than an error is thrown", {
  expect_error(
    country(
      isoCodes = c("FR"),
      services = c("badService")))
})

test_that("If a bad API key is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (apiKey == "BAD_API_KEY") { stop() }
    }, {
      expect_error(
        country(
          isoCodes = c("FR"),
          apiKey = "BAD_API_KEY"))
    }
  )
})

# This test performs real requests to the EPPO API.
test_that("If a bad API key is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    country(
      isoCodes = c("FR"),
      apiKey = "BAD_API_KEY"))
})

test_that("If a bad environment variable is provided than an error is thrown", {
  with_mocked_bindings(
    .fetchService = function(basePath, apiKey, code, service) {
      if (apiKey == '') { stop() }
    }, {
      expect_error(
        country(
          isoCodes = c("FR"),
          apiKey = Sys.getenv("BAD_ENV_VAR")))
    }
  )
})

# This test performs real requests to the EPPO API.
test_that("If a bad environment variable is provided than an error is thrown", {
  skip_on_cran()
  
  expect_error(
    country(
      isoCodes = c("FR"),
      apiKey = Sys.getenv("BAD_ENV_VAR")))
})

#' Query the EPPO REST API and return the result.
#' 
#' This function performs a GET request to the EPPO REST API, automatically
#' builds the full request URL, adds the authentication header, handles
#' HTTP errors, parses the response and adds metadata about the query.
#' 
#' Specific HTTP status codes (400, 401, 403, 404, 429 and 500) are handled
#' explicitly to extract the error message returned by the API; other status
#' codes trigger a connection error.
#' 
#' @param endpoint `character` (string). The relative path of the endpoint to
#'   query, starting with a forward slash (`/`).
#' 
#' @param apiKey `character` (string) The API key used for authentication, sent
#'   via the `X-Api-Key` header.
#' 
#' @param params `list` (optional). A named list of query parameters to include
#'   in the request.
#' 
#' @return A structure containing the parsed JSON response from the API. The
#'   `queried_url` and `queried_on` fields are added to report the queried URL
#'   and the date when the query was performed, respectively.
#' 
#' @details
#' After ensuring that both `endpoint` and `apiKey`are valid strings, the
#' function orchestrates the API call through a set of modular helper functions:
#' 
#' 1. **Request construction** — `.buildRequest()` validates inputs, constructs
#' the full URL, and sets up the GET request with required headers.
#'
#' 2. **Request execution** — `.performRequest()` performs the HTTP call and
#' handles connection-level errors.
#'
#' 3. **HTTP error handling** — `.handleHTTPErrors()` checks the status code
#' and, if not `200`, parses and reports structured error messages returned by
#' the EPPO API.
#'
#' 4. **Response parsing** — `.parseResponse()` converts the successful JSON
#' body.
#' 
#' 5. **Response enriching** - `.enrichResponse()` appends query metadata to the
#' parsed response data.
#'
#' Only the following HTTP status codes are explicitly handled and reported:
#' `400`, `401`, `403`, `404`, `429` and `500`. Any other unexpected code or
#' connection issue triggers an informative error.
#' 
#' @importFrom checkmate assert_string assert_list
#' 
#' @keywords internal
#' @noRd
#' 
.query <- function(endpoint, apiKey, params = list()) {
  
  assert_string(endpoint, pattern = "^/")
  assert_string(apiKey)
  assert_list(params)
  
  baseURL_ <- "https://api.eppo.int/gd/v2"
  fullURL_ <- paste0(baseURL_, endpoint)
  
  request_ <- .buildRequest(
    url = fullURL_,
    apiKey = apiKey,
    doNotFailOn = c(200, 400, 401, 403, 404, 429, 500),
    params = params)
  
  response_ <- .performRequest(request = request_)
  
  .handleHTTPErrors(response = response_)
  
  responseData_ <- .parseResponse(response = response_)
  
  enrichedData_ <- .enrichResponse(responseData = responseData_, url = fullURL_)
  
  return(enrichedData_)
}

test_that("The endpoint must be a string", {
  expect_error(
    .query(
      endpoint = 123,
      apiKey = "EPPO_API_KEY",
      params = list()))
})

test_that("The parameters must be specified in a list", {
  expect_error(
    .query(
      endpoint = "/taxons/taxon/BEMITA/overview",
      apiKey = "EPPO_API_KEY",
      params = 123))
})

test_that("The API key must be a string", {
  expect_error(
    .query(
      endpoint = "/taxons/taxon/BEMITA/overview",
      apiKey = 123))
})

test_that("The endpoint must start with a slash", {
  expect_error(
    .query(
      endpoint = "taxons/taxon/BEMITA/overview",
      apiKey = Sys.getenv("EPPO_API_KEY")))
})

test_that("Expect no errors if the parameters are correct", {
  response_ <- httr2::response(
    status_code = 200,
    url = "https://api.eppo.int/gd/v2/taxons/taxon/BEMITA/names",
    method = "GET",
    headers = list("Content-Type" = "application/json"),
    body = charToRaw("{\"a\": 1}"))
  
  with_mocked_bindings(
    .performRequest = function(request) {
      return(response_)
    }, {
      expect_no_error(
        .query(
          endpoint = "/taxons/taxon/BEMITA/names",
          apiKey = "EPPO_API_KEY"))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect no errors if the parameters are correct", {
  skip_on_cran()
  
  expect_no_error(
    .query(
      endpoint = "/taxons/taxon/BEMITA/names",
      apiKey = Sys.getenv("EPPO_API_KEY")))
})

test_that("Expect an error if a bad endpoint is specified", {
  with_mocked_bindings(
    .performRequest = function(request) {
      if (request$url == "https://api.eppo.int/gd/v2/badService") { stop() }
    }, {
      expect_error(
        .query(
          endpoint = "/badService",
          apiKey = "EPPO_API_KEY"))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect an error if a bad endpoint is specified", {
  skip_on_cran()
  
  expect_error(
    .query(
      endpoint = "/taxons/taxon/BEMITA/badService",
      apiKey = Sys.getenv("EPPO_API_KEY")))
})

test_that("Expect an error if a bad API key is specified", {
  with_mocked_bindings(
    .performRequest = function(request) {
      if (request$headers$"X-Api-Key" == '') { stop() }
    }, {
      expect_error(
        .query(
          endpoint = "/taxons/taxon/BEMITA/overview",
          apiKey = Sys.getenv("BAD_API_KEY")))
    }
  )
})

# This test requires the EPPO_API_KEY environment variable to be set.
# This test performs real requests to the EPPO API.
test_that("Expect an error if a bad API key is specified", {
  skip_on_cran()
  
  expect_error(
    .query(
      endpoint = "/taxons/taxon/BEMITA/overview",
      apiKey = Sys.getenv("BAD_API_KEY")))
})

#' Convert an element to tibble.
#'
#' This helper function takes an elements and converts it into a tibble. If an
#' element is `NULL`, it is replaced with `NA`.
#'
#' @param value. The element to convert to a tibble.
#'
#' @return A tibble containing the data of the original element.
#' 
#' @importFrom tibble as_tibble
#' 
#' @keywords internal
#' @noRd
#'
.valueToTibble <- function(value) {
  
  value[sapply(value, is.null)] <- NA
  value <- as_tibble(value)
  
  return(value)
}

#' Convert nested lists to tibbles and add parent column.
#'
#' This function takes a named list of lists containing dataframes or objects
#' convertible to tibbles, converts each inner element to a tibble, and adds a
#' column containing the name of the parent list.
#'
#' @param nestedValues `list`. A named list of named lists. Each inner list
#'   should contain dataframes or objects convertible to tibbles.
#' 
#' @param parentColumnName `character` (string). A string specifying the name of
#'   the new column that will contain the name of the parent list.
#' 
#' @return A named list of lists of tibbles, with each tibble including an
#'   additional column that indicates the name of its parent list.
#' 
#' @importFrom checkmate assert_list assert_string
#' @importFrom dplyr mutate relocate
#' @importFrom purrr imap map
#' 
#' @keywords internal
#' @noRd
#'
.valuesToTibblesBatch <- function(nestedValues, parentColumnName) {
  
  assert_list(nestedValues)
  assert_string(parentColumnName)
  
  values_ <- imap(nestedValues, ~ {
    x_ <- .valuesToTibbles(values = .x)
    y_ <- .y
    
    map(x_, ~ .x |>
     mutate(!!parentColumnName := y_) |>
     relocate(!!parentColumnName, .before = 1))
  })
  
  return(values_)
}

test_that("The data parameter must be a list", {
  expect_error(
    .groupByServices(
      datasets = 123))
})

test_that("The result must be a list", {
  nestedList_ <- list(
    Group1 = list(
      a = data.frame(x = 1:2, queried_url = NA, queried_on = NA),
      b = data.frame(y = 3:4, queried_url = NA, queried_on = NA)),
    Group2 = list(
      a = data.frame(x = 5:6, queried_url = NA, queried_on = NA),
      b = data.frame(y = 7:8, queried_url = NA, queried_on = NA))
  )
  
  result_ <- .valuesToTibblesBatch(
    nestedValues = nestedList_,
    parentColumnName = "group")
  
  expect_true(
    is.list(
      .groupByServices(
        datasets = result_)))
})

test_that("The resulting list must contain service names", {
  nestedList_ <- list(
    Group1 = list(
      a = data.frame(x = 1:2, queried_url = NA, queried_on = NA),
      b = data.frame(y = 3:4, queried_url = NA, queried_on = NA)),
    Group2 = list(
      a = data.frame(x = 5:6, queried_url = NA, queried_on = NA),
      b = data.frame(y = 7:8, queried_url = NA, queried_on = NA))
  )
  
  result_ <- .valuesToTibblesBatch(
    nestedValues = nestedList_,
    parentColumnName = "group")
  
  result_ <- .groupByServices(
    datasets = result_)
  
  expect_equal(
    names(
      result_),
    c("a", "b"))
})

test_that("The resulting tibbles must contain the parent list name", {
  nestedList_ <- list(
    Group1 = list(
      a = data.frame(x = 1:2, queried_url = NA, queried_on = NA),
      b = data.frame(y = 3:4, queried_url = NA, queried_on = NA)),
    Group2 = list(
      a = data.frame(x = 5:6, queried_url = NA, queried_on = NA),
      b = data.frame(y = 7:8, queried_url = NA, queried_on = NA))
  )
  
  result_ <- .valuesToTibblesBatch(
    nestedValues = nestedList_,
    parentColumnName = "group")
  
  result_ <- .groupByServices(
    datasets = result_)
  
  expect_true("Group1" %in% result_$a$group)
  expect_true("Group2" %in% result_$a$group)
  expect_true("Group1" %in% result_$b$group)
  expect_true("Group2" %in% result_$b$group)
})

test_that("The data parameter must be a list", {
  expect_error(
    .valuesToTibblesBatch(
      nestedValues = 123,
      parentColumnName = "name"))
})

test_that("The column name must be a string", {
  expect_error(
    .valuesToTibblesBatch(
      nestedValues = list(),
      parentColumnName =123))
})

test_that("The result must be a list", {
  nestedList_ <- list(
    Group1 = list(a = data.frame(x = 1:2), b = data.frame(y = 3:4)),
    Group2 = list(a = data.frame(x = 5:6), b = data.frame(y = 7:8))
  )
  
  expect_true(
    is.list(
      .valuesToTibblesBatch(
        nestedValues = nestedList_,
        parentColumnName = "group")))
})

test_that("The resulting tibbles must contain the parent list name", {
  nestedList_ <- list(
    Group1 = list(a = data.frame(x = 1:2), b = data.frame(y = 3:4)),
    Group2 = list(a = data.frame(x = 5:6), b = data.frame(y = 7:8))
  )
  
  result_ <- .valuesToTibblesBatch(
    nestedValues = nestedList_,
    parentColumnName = "group")
  
  expect_true("Group1" %in% result_$Group1$a$group)
  expect_true("Group1" %in% result_$Group1$a$group)
  expect_true("Group2" %in% result_$Group2$b$group)
  expect_true("Group2" %in% result_$Group2$b$group)
})

#' Validate service names against a set of allowed choices.
#'
#' This function checks whether all elements in `services` are included in the
#' allowed `choices`. If any unsupported services are found, the function aborts
#' with an informative error message.
#'
#' @param services `character` (vector). A vector of service names to validate.
#' 
#' @param choices `character` (vector). A vector containing the allowed service
#'   names.
#'
#' @return Invisibly returns `NULL` if all services are valid. Otherwise it
#'   stops with an error.
#' 
#' @importFrom cli cli_abort
#'
#' @keywords internal
#' @noRd
#' 
.checkServices <- function(services, choices) {
  
  assert_vector(services)
  assert_vector(choices)
  
  invalidServices_ <- setdiff(services, choices)
  
  if (length(invalidServices_) > 0) {
    cli_abort(c(
      "Unsupported services requested.",
      "x" = paste(shQuote(invalidServices_), collapse = ", ")
    ))
  }
}

test_that("Services must be specified in a vector", {
  expect_error(
    .checkServices(
      services = 123,
      choices = c("a", "b", "c")))
})

test_that("Choices must be specified in a vector", {
  expect_error(
    .checkServices(
      services = c("a", "b", "c"),
      choices = 123))
})

test_that("Services must be contained in choices", {
  expect_error(
    .checkServices(
      services = c("a", "b"),
      choices = c("d", "e", "f")))
})

test_that("Excepct no error if services are contained in choices", {
  expect_no_error(
    .checkServices(
      services = c("a", "b"),
      choices = c("a", "b", "c")))
})

#' Combine nested lists by child name
#'
#' This function takes a named list of tibbles and combines the inner elements
#' across all parent lists based on their child names. Each resulting tibble
#' corresponds to a unique child name and contains all rows from the nested
#' lists.
#'
#' @param datasets `list`. A named list of lists, where each inner list contains
#'   tibbles or data frames.
#'
#' @return A named list of tibbles. Each tibble contains all combined rows from
#'   the parent lists.
#' 
#' @details
#' The function also ensures that the `queried_url` and `queried_on`
#' columns are always placed after the others.
#' 
#' @importFrom purrr map imap
#' @importFrom checkmate assert_list
#' @importFrom dplyr bind_rows relocate
#' 
#' @keywords internal
#' @noRd
#'
.groupByServices <- function(datasets) {
  
  assert_list(datasets)
  
  serviceNames_ <- unique(unlist(map(datasets, names)))
  
  datasets <- map(serviceNames_, function(serviceName_) {
    bind_rows(map(datasets, ~ .x[[serviceName_]]))
  })
  
  names(datasets) <- serviceNames_
  
  datasets <- imap(datasets, ~ .x |>
   relocate(queried_on, .after = last_col()) |>
   relocate(queried_url, .after = last_col()))
  
  return(datasets)
}

#' Perform an HTTP request to the with error handling.
#' 
#' This helper function executes a prepared HTTP request and safely handles any
#' connection-level errors. If an error occurs during the request (e.g., network
#' failure, timeout, DNS issue), it aborts execution with a clear message.
#' 
#' @param request `httr2_request`. A request object created with the `httr2`
#'   package and configured with methods, headers and parameters.
#' 
#' @return The HTTP response object returned by `httr2::req_perform()` if the
#'   request is successful.
#' 
#' @importFrom checkmate assert_class
#' @importFrom httr2 req_perform
#' @importFrom cli cli_abort
#' 
#' @keywords internal
#' @noRd
#' 
.performRequest <- function(request) {
  
  assert_class(request, "httr2_request")
  
  response_ <- tryCatch(
    req_perform(request),
    error = function(e) {
      cli_abort(c(
        "HTTP request failed.",
        "x" = conditionMessage(e)
      ))
    }
  )
  
  return(response_)
}

#' Convert list elements to tibbles.
#'
#' This helper function takes a list of elements and converts each element into
#' a tibble.
#'
#' @param values `list`. The list whose elements must be converted to tibbles.
#'
#' @return A list where each original element has been converted to a tibble.
#'
#' @importFrom checkmate assert_list
#' @importFrom purrr map
#' 
#' @keywords internal
#' @noRd
#'
.valuesToTibbles <- function(values) {
  
  assert_list(values)
  
  values <- map(values, ~ .valueToTibble(value = .x))
  
  return(values)
}

test_that("The data parameter must be a list", {
  expect_error(
    .valuesToTibbles(
      values = 123))
})

test_that("The result must be a list", {
  expect_true(
    is.list(
      .valuesToTibbles(
        values = list(
          a = list(x = 1:3, y = 4:6),
          b = list(x = 7:9, y = 10:12)))))
})

test_that("All the elements must be converted to a tibble", {
  data_ <- .valuesToTibbles(
    values = list(
      a = list(x = 1:3, y = 4:6),
      b = list(x = 7:9, y = 10:12)))
  
  expect_true(
    is.data.frame(
      data_$a))
  
  expect_true(
    is.data.frame(
      data_$b))
})

#' Query the EPPO API References endpoints.
#' 
#' This function queries the References endpoints of the EPPO Global Database
#' via REST API. The function sequentially queries all specified `services` and
#' returns the extracted data through a list of dataframes.
#'
#' @param services `character` (vector). One or more References services to
#'   query. Common service names include: `rppos`, `qList`,
#'   `distributionStatus`, `pestHostClassification`, `vectorClassification`,
#'   `countries` and `countriesStates`. A validation step ensures that all
#'   provided services match the supported service names.
#' 
#'   By default: all services.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content.
#' 
#' @importFrom checkmate assert_vector assert_string
#' @importFrom purrr map set_names
#'
#' @examples
#' \dontrun{
#' # Get all information.
#' referencesData_ <- references()
#' 
#' # Get information about distribution status codes.
#' referencesData_ <- references(services = c("distributionStatus"))
#' 
#' # Get information about EPPO list codes and labels and countries.
#' referencesData_ <- references(services = c("qList", "countries"))
#' }
#' 
#' @export
#' 
references <- function(
  services = c(
    "rppos",
    "qList",
    "distributionStatus",
    "pestHostClassification",
    "vectorClassification",
    "countries",
    "countriesStates"),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(services)
  assert_string(apiKey)
  
  .checkServices(services, eval(formals()$services))
  
  referencesData_ <- map(services, ~ .fetchService(
    basePath = "/references",
    apiKey = apiKey,
    service = .x)) |>
    set_names(services)
  
  referencesData_ <- .rebuildReferences(referencesData = referencesData_)
  
  return(referencesData_)
}

#' Build the EPPO reporting-related service path.
#'
#' This helper function constructs the endpoint path for EPPO API reporting
#' or article services based on the service type and relevant identifiers
#' provided in the `params` list. It also handles missing or invalid parameters
#' by raising informative errors.
#' 
#' @param service `character` (string). The type of service. Supported values
#'   are `reporting` and `article`.
#' 
#' @param params `list`. A named list containing identifiers needed for the
#'   endpoint path. Must include `reporting_id` if service is "reporting" or
#'   `article_id` if service is "article".
#' 
#' @return The constructed service path to use with API calls.
#' 
#' @details
#' The function checks for the presence of the required identifiers in `params`.
#' If a required parameter is missing or invalid, an error is thrown, with a
#' message indicating the cause of the failure.
#' 
#' @importFrom checkmate assert_string assert_list
#' @importFrom glue glue
#'
#' @keywords internal
#' @noRd
#' 
.buildReportingServicePath <- function(service, params) {
  
  assert_string(service)
  assert_list(params)
  
  requiredId_ <- list(
    reporting = "reporting_id",
    article = "article_id"
  )
    
  requiredIdName_ <- requiredId_[[service]]
  
  if (!is.null(requiredIdName_)) {
    if (is.null(params[[requiredIdName_]])) {
      stop(glue("Missing parameter: `{requiredIdName_}`"))
    }
    
    service <- glue("{service}/{params[[requiredIdName_]]}")
  }
  
  return(service)
}

test_that("The service must be a string", {
  expect_error(
    .buildReportingServicePath(
      service = 123,
      params = list()))
})

test_that("The parameters must be specified in a list", {
  expect_error(
    .buildReportingServicePath(
      service = "article",
      params = 123))
})

test_that("The result must be a string", {
  expect_true(
    is.character(
      .buildReportingServicePath(
        service = "article",
        params = list(
          article_id = 10))))
})

test_that("The result must be correct for services with no parameters", {
  expect_equal(
    .buildReportingServicePath(
      service = "list",
      params = list()),
    "list")
})

test_that("The result must be correct for services with parameters", {
  expect_equal(
    .buildReportingServicePath(
      service = "article",
      params = list(
        article_id = 10)),
    "article/10")
})

test_that("Expect an error if required parameters are missing", {
  expect_error(
    .buildReportingServicePath(
      service = "reporting",
      params = list(
        wrong_param = 10)))
  
  expect_error(
    .buildReportingServicePath(
      service = "article",
      params = list(
        wrong_param = 10)))
})

#' Handle non-successful HTTP responses from the EPPO API.
#' 
#' This helper function checks whether an HTTP response from the EPPO API
#' indicates success (status code 200). If the response contains any other
#' status code, it attempts to parse the JSON body for an error message and
#' raises a formatted error.
#' 
#' @param response `httr2_response`. An object returned by `req_perform()`.
#' 
#' @return Invisibly returns `NULL` if the response is successful (status code
#'   200). Otherwise, execution stops with a descriptive error message.
#' 
#' @details
#' The function is designed to simplify EPPO API error handling by centralizing
#' logic for non-200 status codes. It gracefully handles malformed or non-JSON
#' responses by returning `NULL` if parsing fails.
#' 
#' @importFrom checkmate assert_class
#' @importFrom httr2 resp_status resp_body_json
#' @importFrom cli cli_abort
#' 
#' @keywords internal
#' @noRd
#' 
.handleHTTPErrors <- function(response) {
  
  assert_class(response, "httr2_response")
  
  if (resp_status(response) != 200) {
    responseJSONBody_ <- tryCatch(
      resp_body_json(response),
      error = function(e) {
        return(NULL)
      }
    )
    
    cli_abort(c(
      "API request failed.",
      "x" = responseJSONBody_$error
    ))
  }
}

#' Query the EPPO API Taxons endpoints.
#' 
#' This function queries the Taxons endpoints of the EPPO Global Database via
#' REST API. The function sequentially queries all specified `services` and
#' returns the extracted data through a list of dataframes.
#'
#' @param services `character` (vector). One or more Taxons services to query.
#'   Common service names include: `list`. A validation step ensures that all
#'   provided services match the supported service names.
#' 
#'   By default: all services.
#' 
#' @param params `list` (optional). A named list of query parameters to include
#'   in the request. The list of available parameters can be accessed via the
#'   [EPPO API Documentation](https://data.eppo.int/ui/#/docs/GDAPI)
#'   platform.
#' 
#'   By default: default parameters.
#'
#' @param apiKey `character` (string). The API key used for authentication. It
#'   can be specified manually or through the .Renviron file.
#' 
#'   By default: Sys.getenv("EPPO_API_KEY").
#'
#' @return A named list of results, in which each level corresponds to the data
#'   retrieved for each specified service. Each service element contains a
#'   flattened dataframe with the queried content.
#' 
#' @importFrom checkmate assert_vector assert_list assert_string assert
#' @importFrom purrr map2 set_names
#'
#' @examples
#' \dontrun{
#' # Get the list of taxons.
#' taxonsData_ <- taxons()
#' 
#' # Get the list of taxons with custom parameters.
#' taxonsData_ <- taxons(params = list(
#'   "list" = list(
#'     createdFromDate = "2000-01-01",
#'     limit = 5,
#'     offset = 1,
#'     orderAsc = FALSE,
#'     orderBy = "eppocode")))
#' }
#' 
#' @export
#' 
taxons <- function(
  services = c(
    "list"),
  params = list(
    "list" = list(
      limit = 100,
      offset = 0,
      orderAsc = TRUE,
      orderBy = "creation"
    )),
  apiKey = Sys.getenv("EPPO_API_KEY")) {
  
  assert_vector(services)
  assert_list(params)
  assert_string(apiKey)
  assert(length(services) == length(params))
  
  .checkServices(services, eval(formals()$services))
  
  taxonsData_ <- map2(services, params[services], ~ .fetchService(
    basePath = "/taxons",
    apiKey = apiKey,
    service = .x,
    params = .y)) |>
    set_names(services)
  
  taxonsData_ <- .rebuildTaxons(taxonsData = taxonsData_)
  
  return(taxonsData_)
}

---
title: "eppoFindeR"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{eppoFindeR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Overview

**eppoFindeR** provides an R interface to the public APIs of the **European and Mediterranean Plant Protection Organization (EPPO)** database. The package facilitates access to a wide range of pest-related information collected and maintained by EPPO, allowing users to query, retrieve, and process this data directly from R.

The package is intended for researchers, analysts, and practitioners in plant protection who require convenient programmatic access to EPPO data.

## Installation

### From CRAN

```{r eval=FALSE}
install.packages("eppoFindeR")
```

### Development version (from GitHub)

To install the latest development version:

```{r eval=FALSE}
# install.packages("devtools")
devtools::install_github("openefsa/eppoFindeR")
```

## Requirements

An active internet connection is required, as the package communicates with EPPO’s online services to fetch and process data.

## Working with API keys

The *eppoFindeR* package requires your unique API token to function properly. You can provide this token in one of two ways:

1.  By setting it in the `.Renviron` file.\
2.  By including it manually in each API request.

### Setting the API key via `.Renviron`

The `.Renviron` file is used to define environment variables that R loads automatically at the start of each session. This approach is particularly convenient for sensitive information like API keys, as it allows you to use them in any R script or function without hardcoding them.

Place the `.Renviron` file in your home directory (for example, `C:/Users/username/Documents/.Renviron` on Windows or `~/.Renviron` on Unix-like systems). You can create or edit this file with any plain text editor.

Add your EPPO API key in the following format:

`EPPO_API_KEY=<your_eppo_api_key>`

After saving the file, R will automatically read the token on startup.

### Setting the API key manually for each request

Alternatively, you can provide the API key directly in the `apiKey` argument of functions that require it. This is useful if you want to use different tokens for different calls, or if you prefer not to store the token globally. For example:

```{r apiKeyExample, eval=FALSE}
taxon(c("BEMITA"), apiKey = "<your_eppo_api_key>")
```

## Basic usage

The main purpose of *eppoFindeR* is to query the EPPO database for specific EPPO codes and retrieve relevant information across various endpoints. For each endpoint, you can either:

1.  Query all available services to obtain more comprehensive data.\
2.  Query one or more services to obtain specific data.

For plant and pest species, the basic information returned includes scientific names, synonyms, common names in different languages, and taxonomic classification. For pests of regulatory interest, you can also retrieve more detailed information, such as host plants and quarantine categorization.

Below are examples demonstrating how to use the functions in this package. First, load the *eppoFindeR* package:

```{r loadLibrary, eval=FALSE}
library(eppoFindeR)
```

To explore the arguments and usage of a specific function, you can run:

```{r eval=FALSE}
help("<function_name>")
```

This will show the full documentation for the function, including its arguments, return values, and usage examples.

For example, if you are working with the `taxon()` function, you can check its documentation with:

```{r eval=FALSE}
help("taxon")
```

## Querying a specific category

The *eppoFindeR* package allows you to query all categories available in version 2.0 of the EPPO APIs: General, Taxons, Taxon, Country, RPPO, Tools, Reporting Service, and References.

Each category has a corresponding function in the package with the same name in *lowerCamelCase* format: general(), taxons(), taxon(), country(), rppo(), tools(), reportingService(), and references(). By default, these functions return all data available under the selected category, but you can customize the query by specifying the desired services.

For example, to query all services of the Taxon category for the EPPO code "BEMITA", you can use the following code:

```{r taxonAllServices, eval=FALSE}
taxon(c("BEMITA"))
```

Some services require certain mandatory parameters, which must be provided when making the request. For example:

```{r toolsName2Codes, eval=FALSE}
tools(
  c("name2codes"), # The service to query (/tools/name2codes).
  params = list(
    name2codes = list(
      name = "Bemisia tabaci" # The required parameter.
    )
  )
)
```

## Querying a specific service

To query a specific service within an endpoint, eppoFindeR allows you to specify it directly in the function call. For example, to access only the `/taxons/taxon/categorization` service for the EPPO code "BEMITA", you can use the `taxon()` function as follows:

```{r queryService, eval=FALSE}
taxon(c("BEMITA"), c("categorization"))
```

### Expected output

```{r queryServiceOutput, eval=FALSE}
$categorization
# A tibble: 20 × 12
   queried_eppo_code continent_id continent_name country_iso country_name    qlist qlist_label year_add
   <chr>                    <int> <chr>          <chr>       <chr>           <chr> <chr>          <int>
 1 BEMITA                       4 America        CL          Chile           1     A1 list         2019
 2 BEMITA                       2 Asia           BH          Bahrain         2     A2 list         2003
 3 BEMITA                       2 Asia           KZ          Kazakhstan      1     A1 list         2017
 4 BEMITA                       1 Europe         AZ          Azerbaijan      1     A1 list         2024
 5 BEMITA                       1 Europe         BY          Belarus         X     Quarantine…     1994
 6 BEMITA                       1 Europe         GE          Georgia         1     A1 list         2018
 7 BEMITA                       1 Europe         MD          Moldova, Repub… X     Quarantine…     2017
 8 BEMITA                       1 Europe         NO          Norway          X     Quarantine…     2012
 9 BEMITA                       1 Europe         RU          Russian Federa… 2     A2 list         2014
10 BEMITA                       1 Europe         RS          Serbia          1     A1 list         2015
11 BEMITA                       1 Europe         CH          Switzerland     1     A1 list         2019
12 BEMITA                       1 Europe         TR          Türkiye         2     A2 list         2016
13 BEMITA                       1 Europe         UA          Ukraine         1     A1 list         2019
14 BEMITA                       1 Europe         GB          United Kingdom  1     A1 list         2020
15 BEMITA                       5 Oceania        NZ          New Zealand     X     Quarantine…     2000
16 BEMITA                       6 RPPO/EU        9M          EAEU            2     A2 list         2016
17 BEMITA                       6 RPPO/EU        9A          EPPO            2     A2 list         1989
18 BEMITA                       6 RPPO/EU        9L          EU              P     PZ Quarant…     2019
19 BEMITA                       6 RPPO/EU        9L          EU              N     A1 Quarant…     2019
20 BEMITA                       6 RPPO/EU        9H          OIRSA           2     A2 list         1992
# ℹ 4 more variables: year_delete <lgl>, year_transient <lgl>, queried_url <chr>, queried_on <dttm>
```

All function calls return a list of dataframes, each containing the information from the requested services.

## Querying all available services

To fetch data from all available services within a specific endpoint, simply use the corresponding function without modifying the services parameter. Each function will return a list of dataframes, with each dataframe containing information from one service.

```{r queryWhole, eval=FALSE}
data <- taxon(c("BEMITA"))
```

### Expected output

```{r queryAllOutput, eval=FALSE}
# Print the names of the services in the result.
names(data)
 [1] "overview"           "infos"              "names"              "taxonomy"
 [5] "categorization"     "kingdom"            "hosts"              "pests"           
 [9] "vectors"            "vectorof"           "bca"                "bcaof"           
[13] "photos"             "reporting_articles" "documents"          "standards"       
[17] "distribution"      

# Print the data from a specific service.
data$overview
# A tibble: 1 × 11
  queried_eppo_code eppocode datecreate         lastupdate infos replacedby prefname is_active datatype
  <chr>             <chr>    <chr>              <chr>      <lgl> <lgl>      <chr>    <lgl>     <chr>   
1 BEMITA            BEMITA   2002-10-28 00:00:… 2002-10-2… NA    NA         Bemisia… TRUE      GAI     
# ℹ 2 more variables: queried_url <chr>, queried_on <dttm>
```

Using `names(data)` will display the names of the queried services. You can then access a specific service to view the data associated with it.

## Data wrangling

The *eppoFindeR* package provides a convenient function, `uniformTaxonomy()`, to create a complete and standardized taxonomy dataframe. This function takes the taxonomy data returned by the taxonomy service and ensures a uniform structure, including all expected taxonomic ranks, even if some ranks are missing in the original result.

For example, assume you have retrieved taxonomy data for the EPPO code "BEMITA" using the `taxon()` function:

```{r eval=FALSE}
data <- taxon(c("BEMITA"), c("taxonomy"))
```

You can then generate a uniform taxonomy table with all ranks using:

```{r uniformTaxonomy, eval=FALSE}
taxonomy <- uniformTaxonomy(data$taxonomy)
```

### Expected output

```{r uniformTaxonomyOutput, eval=FALSE}
# A tibble: 11 × 6
   type      queried_eppo_code eppocode prefname       queried_url                  queried_on         
   <chr>     <chr>             <chr>    <chr>          <chr>                        <dttm>             
 1 Kingdom   BEMITA            1ANIMK   Animalia       https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 2 Phylum    BEMITA            1ARTHP   Arthropoda     https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 3 Subphylum BEMITA            1HEXAQ   Hexapoda       https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 4 Class     BEMITA            1INSEC   Insecta        https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 5 Subclass  NA                NA       NA             NA                           NA                 
 6 Order     BEMITA            1HEMIO   Hemiptera      https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 7 Suborder  BEMITA            1STERR   Sternorrhyncha https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 8 Family    BEMITA            1ALEYF   Aleyrodidae    https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
 9 Subfamily NA                NA       NA             NA                           NA                 
10 Genus     BEMITA            1BEMIG   Bemisia        https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
11 Species   BEMITA            BEMITA   Bemisia tabaci https://api.eppo.int/gd/v2/… 2025-11-18 14:29:50
```

## Data retrieval for multiple EPPO or ISO codes

The `taxon()` and `country()` functions allow you to retrieve also a complete data dump for multiple EPPO or ISO codes by querying all available EPPO services for each of them.

Here is an example showing how to use the function to fetch all available information for the EPPO codes "BEMITA" and "GOSHI":

```{r multipleEPPOCodes, eval=FALSE}
data <- taxon(c("BEMITA", "GOSHI"))
```

### Expected output

```{r multipleEPPOCodesOutput, eval=FALSE}
# Print the names of the services in the result.
names(data)
 [1] "overview"           "infos"              "names"              "taxonomy"
 [5] "categorization"     "kingdom"            "hosts"              "pests"           
 [9] "vectors"            "vectorof"           "bca"                "bcaof"           
[13] "photos"             "reporting_articles" "documents"          "standards"       
[17] "distribution"      

# Print the data from a specific service.
data$overview
# A tibble: 2 × 11
  queried_eppo_code eppocode datecreate         lastupdate infos replacedby prefname is_active datatype
  <chr>             <chr>    <chr>              <chr>      <chr> <lgl>      <chr>    <lgl>     <chr>   
1 BEMITA            BEMITA   2002-10-28 00:00:… 2002-10-2… NA    NA         Bemisia… TRUE      GAI     
2 GOSHI             GOSHI    2004-04-21 00:00:… 2004-04-2… <p>M… NA         Gossypi… TRUE      PFL     
# ℹ 2 more variables: queried_url <chr>, queried_on <dttm>
```

Using `names(data)` will display the names of the queried services. You can then access a specific service to view the data associated with it.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rppo.R
\name{rppo}
\alias{rppo}
\title{Query the EPPO API RPPO endpoints.}
\usage{
rppo(
  rppoCodes,
  services = c("overview", "categorization"),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{rppoCodes}{\code{character} (vector). One or more RPPO codes to query.
Available RPPO codes can be retrieved via the \code{references/rppos} service.}

\item{services}{\code{character} (vector). One or more RPPO services to query.
Common service names include: \code{overview} and \code{categorization}. A
validation step ensures that all provided services match the supported
service names.

By default: all services.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content for all the specified RPPO
codes.
}
\description{
This function queries the RPPO endpoints of the EPPO Global Database via
REST API for one or more RPPO code(s) and one or more service(s). For each
RPPO code in \code{rppoCodes}, the function sequentially queries all specified
\code{services} and returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get all information about the European and Mediterranean Plant Protection
# Organisation.
rppoData_ <- rppo(rppoCodes = c("9A"))

# Get categorization data about the European and Mediterranean Plant
# Protection Organisation.
rppoData_ <- rppo(rppoCodes = c("9A"), services = c("categorization"))

# Get basic data about European and Mediterranean Plant Protection
# Organisation and the European Union.
rppoData_ <- rppo(rppoCodes = c("9A", "9L"), services = c("overview"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eppoFindeR-package.R
\docType{package}
\name{eppoFindeR-package}
\alias{eppoFindeR}
\alias{eppoFindeR-package}
\title{eppoFindeR: Interface to the EPPO Database and Public APIs}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides an interface to the public APIs of the European and Mediterranean Plant Protection Organization (EPPO) database. It enables users to retrieve EPPO data by accessing specific services and datasets. The package also includes utilities for data wrangling, including the integration of taxonomy with rank information. See \url{https://data.eppo.int/apis/} for more details.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://openefsa.github.io/eppoFindeR/}
  \item Report bugs at \url{https://github.com/openefsa/eppoFindeR/issues}
}

}
\author{
\strong{Maintainer}: Luca Belmonte \email{luca.belmonte@efsa.europa.eu} (\href{https://orcid.org/0000-0002-7977-9170}{ORCID})

Authors:
\itemize{
  \item Lorenzo Copelli (\href{https://orcid.org/0009-0002-4305-065X}{ORCID})
  \item Dayana Stephanie Buzle (\href{https://orcid.org/0009-0003-2990-7431}{ORCID})
  \item Rafael Vieira (\href{https://orcid.org/0009-0009-0289-5438}{ORCID})
  \item Agata Kaczmarek (\href{https://orcid.org/0000-0002-7463-5821}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uniformTaxonomy.R
\name{uniformTaxonomy}
\alias{uniformTaxonomy}
\title{Create a complete and uniform taxonomy dataframe.}
\usage{
uniformTaxonomy(taxonomyData)
}
\arguments{
\item{taxonomyData}{\code{tibble}. A dataframe containing taxonomy data provided
by the EPPO service for a given EPPO code.}
}
\value{
A dataframe where each row represents one of the expected taxonomic
ranks. Fields corresponding to ranks not present in the original taxonomy
are filled with \code{NA}. The \code{level} column is excluded from the output. The
column \code{queried_eppo_code} NA values are filled with the right code.
}
\description{
This function normalizes the taxonomy returned by the EPPO service,
producing a uniform structure that includes all possible taxonomic
categories, even when some of them are not present in the original result.
}
\details{
The function uses an internal set of known taxonomic ranks and performs
a left join with the \code{taxonomy} dataframe, filling missing ranks with \code{NA}.
After the join, the column \code{level} from the original taxonomy is removed,
since the rank order is fully defined by the internal reference list. The
column \code{queried_eppo_code} NA values are filled with the right code.
}
\examples{
\dontrun{
# Retrieve taxonomy data from the EPPO service.
taxonData_ <- taxon(eppoCodes = c("BEMITA"), services = c("taxonomy"))

# Create a uniform taxonomy with all ranks.
taxonomyData_ <- uniformTaxonomy(taxonomyData = taxonData_$taxonomy)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/general.R
\name{general}
\alias{general}
\title{Query the EPPO API General endpoint.}
\usage{
general(services = c("status"), apiKey = Sys.getenv("EPPO_API_KEY"))
}
\arguments{
\item{services}{\code{character} (vector). One or more General services to query.
Common service names include: \code{status}. A validation step ensures that all
provided services match the supported service names.

By default: all services.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content.
}
\description{
This function queries the General endpoints of the EPPO Global Database via
REST API. The function sequentially queries all specified \code{services} and
returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get information about system health status.
generalData_ <- general()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxons.R
\name{taxons}
\alias{taxons}
\title{Query the EPPO API Taxons endpoints.}
\usage{
taxons(
  services = c("list"),
  params = list(list = list(limit = 100, offset = 0, orderAsc = TRUE, orderBy =
    "creation")),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{services}{\code{character} (vector). One or more Taxons services to query.
Common service names include: \code{list}. A validation step ensures that all
provided services match the supported service names.

By default: all services.}

\item{params}{\code{list} (optional). A named list of query parameters to include
in the request. The list of available parameters can be accessed via the
\href{https://data.eppo.int/ui/#/docs/GDAPI}{EPPO API Documentation}
platform.

By default: default parameters.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content.
}
\description{
This function queries the Taxons endpoints of the EPPO Global Database via
REST API. The function sequentially queries all specified \code{services} and
returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get the list of taxons.
taxonsData_ <- taxons()

# Get the list of taxons with custom parameters.
taxonsData_ <- taxons(params = list(
  "list" = list(
    createdFromDate = "2000-01-01",
    limit = 5,
    offset = 1,
    orderAsc = FALSE,
    orderBy = "eppocode")))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxon.R
\name{taxon}
\alias{taxon}
\title{Query the EPPO API Taxon endpoints.}
\usage{
taxon(
  eppoCodes,
  services = c("overview", "infos", "names", "taxonomy", "categorization", "kingdom",
    "hosts", "pests", "vectors", "vectorof", "bca", "bcaof", "photos",
    "reporting_articles", "documents", "standards", "distribution"),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{eppoCodes}{\code{character} (vector). One or more EPPO codes to query.

EPPO codes are standardized identifiers for plants and pests (including
pathogens). These alphanumeric codes, typically of 5 to 6 letters long,
serve as unique, mnemonic abbreviations of scientific names. The EPPO
coding system facilitates the management and exchange of organism names
across databases and IT systems. EPPO codes and their corresponding data
can be accessed via the \href{https://data.eppo.int/}{EPPO Data Services}
platform.}

\item{services}{\code{character} (vector). One or more Taxon services to query.
Common service names include: \code{overview}, \code{infos}, \code{names}, \code{taxonomy},
\code{categorization}, \code{kingdom}, \code{hosts}, \code{pests}, \code{vectors}, \code{vectorof},
\code{bca}, \code{bcaof}, \code{photos}, \code{reporting_articles}, \code{documents}, \code{standards}
and \code{distribution}. A validation step ensures that all provided services
match the supported service names.

By default: all services.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content for all the specified EPPO
codes.
}
\description{
This function queries the Taxon endpoints of the EPPO Global Database via
REST API for one or more EPPO code(s) and one or more service(s). For each
EPPO code in \code{eppoCodes}, the function sequentially queries all specified
\code{services} and returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get all information about Bemisia tabaci.
taxonData_ <- taxon(eppoCodes = c("BEMITA"))

# Get names data about Bemisia tabaci.
taxonData_ <- taxon(eppoCodes = c("BEMITA"), services = c("names"))

# Get taxonomy and categorization data about Bemisia tabaci and Gossypium
# hirsutum.
taxonData_ <- taxon(
  eppoCodes = c("BEMITA", "GOSHI"),
  services = c("taxonomy", "categorization"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/references.R
\name{references}
\alias{references}
\title{Query the EPPO API References endpoints.}
\usage{
references(
  services = c("rppos", "qList", "distributionStatus", "pestHostClassification",
    "vectorClassification", "countries", "countriesStates"),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{services}{\code{character} (vector). One or more References services to
query. Common service names include: \code{rppos}, \code{qList},
\code{distributionStatus}, \code{pestHostClassification}, \code{vectorClassification},
\code{countries} and \code{countriesStates}. A validation step ensures that all
provided services match the supported service names.

By default: all services.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content.
}
\description{
This function queries the References endpoints of the EPPO Global Database
via REST API. The function sequentially queries all specified \code{services} and
returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get all information.
referencesData_ <- references()

# Get information about distribution status codes.
referencesData_ <- references(services = c("distributionStatus"))

# Get information about EPPO list codes and labels and countries.
referencesData_ <- references(services = c("qList", "countries"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reportingService.R
\name{reportingService}
\alias{reportingService}
\title{Query the EPPO API Taxons endpoints.}
\usage{
reportingService(
  services = c("list", "reporting", "article"),
  params = list(list = list(), reporting = list(), article = list()),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{services}{\code{character} (vector). One or more Taxons services to query.
Common service names include: \code{list}, \code{reporting} and \code{article}. A
validation step ensures that all provided services match the supported
service names.

By default: all services.}

\item{params}{\code{list}. A named list of query parameters to include in the
request. The list of available parameters can be accessed via the
\href{https://data.eppo.int/ui/#/docs/GDAPI}{EPPO API Documentation}
platform.

By default: no parameters, you need to specify the required ones.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content.
}
\description{
This function queries the Taxons endpoints of the EPPO Global Database via
REST API. The function sequentially queries all specified \code{services} and
returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get the list of reporting service issues.
reportingServiceData_ <- reportingService(
  services = c("list"),
  params = list(
    "list" = list()))

# Get a specific reporting service issue by ID.
reportingServiceData_ <- reportingService(
  services = c("reporting"),
  params = list(
    reporting = list(
      reporting_id = 10)))

# Get a specific article by ID.
reportingServiceData_ <- reportingService(
  services = c("article"),
  params = list(
    article = list(
      article_id = 234)))

# Get a specific reporting service issue and a specific article by ID.
reportingServiceData_ <- reportingService(
  services = c("reporting", "article"),
  params = list(
    reporting = list(reporting_id = 10),
    article = list(article_id = 234)
  ))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/country.R
\name{country}
\alias{country}
\title{Query the EPPO API Country endpoints.}
\usage{
country(
  isoCodes,
  services = c("overview", "categorization", "presence"),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{isoCodes}{\code{character} (vector). One or more ISO codes to query.

ISO country codes are standardized identifiers for countries and
territories. These codes, typically consisting of two-letter, provide
unique and consistent abbreviations of country names. The ISO coding system
enables easy management, exchange, and integration of country information
across databases and IT systems. ISO country codes and their related data
can be accessed via the
\href{https://www.iso.org/obp/ui/#search}{ISO Online Browsing Platform}.}

\item{services}{\code{character} (vector). One or more Country services to query.
Common service names include: \code{overview}, \code{categorization} and \code{presence}.
A validation step ensures that all provided services match the supported
service names.

By default: all services.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content for all the specified ISO
codes.
}
\description{
This function queries the Country endpoints of the EPPO Global Database via
REST API for one or more ISO code(s) and one or more service(s). For each
ISO code in \code{isoCodes}, the function sequentially queries all specified
\code{services} and returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get all information about France.
countryData_ <- country(isoCodes = c("FR"))

# Get categorization data about France.
countryData_ <- country(isoCodes = c("FR"), services = c("categorization"))

# Get basic data about France and Italy.
countryData_ <- country(isoCodes = c("FR", "IT"), services = c("overview"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{tools}
\alias{tools}
\title{Query the EPPO API Tools endpoints.}
\usage{
tools(
  services = c("name2codes"),
  params = list(name2codes = list()),
  apiKey = Sys.getenv("EPPO_API_KEY")
)
}
\arguments{
\item{services}{\code{character} (vector). One or more Tools services to query.
Common service names include: \code{name2codes}. A validation step ensures that
all provided services match the supported service names.

By default: all services.}

\item{params}{\code{list}. A named list of query parameters to include in the
request. The list of available parameters can be accessed via the
\href{https://data.eppo.int/ui/#/docs/GDAPI}{EPPO API Documentation}
platform.

By default: no parameters, you need to specify the required ones.}

\item{apiKey}{\code{character} (string). The API key used for authentication. It
can be specified manually or through the .Renviron file.

By default: Sys.getenv("EPPO_API_KEY").}
}
\value{
A named list of results, in which each level corresponds to the data
retrieved for each specified service. Each service element contains a
flattened dataframe with the queried content.
}
\description{
This function queries the Tools endpoints of the EPPO Global Database via
REST API. The function sequentially queries all specified \code{services} and
returns the extracted data through a list of dataframes.
}
\examples{
\dontrun{
# Get the EPPO codes associated to the name Bemisia tabaci.
toolsData_ <- tools(params = list(
  name2codes = list(
    onlyPreferred = FALSE,
    name = "Bemisia tabaci")))
}

}
